from typing import AbstractSet, Literal, Union

RecordingMode = Literal['standard', 'raw']

TextMessageType = Literal['text']
HtmlMessageType = Literal['html']
MarkdownMessageType = Literal['markdown']
MessageType = Union[TextMessageType, MarkdownMessageType, HtmlMessageType]

EmailMessageType = Union[TextMessageType, HtmlMessageType]
ServerchanMessageType = MarkdownMessageType
PushdeerMessageType = Union[TextMessageType, MarkdownMessageType]
PushplusMessageType = Union[TextMessageType, MarkdownMessageType, HtmlMessageType]
TelegramMessageType = Union[MarkdownMessageType, HtmlMessageType]
BarkMessageType = Union[TextMessageType, MarkdownMessageType]


KeyOfSettings = Literal[
    'version',
    'tasks',
    'output',
    'logging',
    'bili_api',
    'header',
    'danmaku',
    'recorder',
    'postprocessing',
    'space',
    'email_notification',
    'serverchan_notification',
    'pushdeer_notification',
    'pushplus_notification',
    'telegram_notification',
    'bark_notification',
    'webhooks',
]

KeySetOfSettings = AbstractSet[KeyOfSettings]



from __future__ import annotations

import logging
from typing import Optional
from urllib.parse import urlparse

import requests
from reactivex import Observable, abc
from reactivex import operators as ops

from blrec.bili.exceptions import (
    LiveRoomEncrypted,
    LiveRoomHidden,
    LiveRoomLocked,
    NoAlternativeStreamAvailable,
    NoStreamAvailable,
    NoStreamCodecAvailable,
    NoStreamFormatAvailable,
    NoStreamQualityAvailable,
)
from blrec.bili.live import Live
from blrec.utils import operators as utils_ops
from blrec.utils.mixins import AsyncCooperationMixin

from ..stream_param_holder import StreamParamHolder, StreamParams

__all__ = ('StreamURLResolver',)


logger = logging.getLogger(__name__)


class StreamURLResolver(AsyncCooperationMixin):
    def __init__(self, live: Live, stream_param_holder: StreamParamHolder) -> None:
        super().__init__()
        self._live = live
        self._stream_param_holder = stream_param_holder
        self._stream_url: str = ''
        self._stream_host: str = ''
        self._stream_params: Optional[StreamParams] = None

    @property
    def stream_url(self) -> str:
        return self._stream_url

    @property
    def stream_host(self) -> str:
        return self._stream_host

    def reset(self) -> None:
        self._stream_url = ''
        self._stream_host = ''
        self._stream_params = None

    def __call__(self, source: Observable[StreamParams]) -> Observable[str]:
        self.reset()
        return self._solve(source).pipe(  # type: ignore
            ops.do_action(on_error=self._before_retry),
            utils_ops.retry(delay=1, should_retry=self._should_retry),
        )

    def _solve(self, source: Observable[StreamParams]) -> Observable[str]:
        def subscribe(
            observer: abc.ObserverBase[str],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            def on_next(params: StreamParams) -> None:
                if self._can_resue_url(params):
                    observer.on_next(self._stream_url)
                    return

                try:
                    logger.info(
                        f'Getting the live stream url... '
                        f'qn: {params.quality_number}, '
                        f'format: {params.stream_format}, '
                        f'api platform: {params.api_platform}, '
                        f'use alternative stream: {params.use_alternative_stream}'
                    )
                    url = self._run_coroutine(
                        self._live.get_live_stream_url(
                            params.quality_number,
                            api_platform=params.api_platform,
                            stream_format=params.stream_format,
                            select_alternative=params.use_alternative_stream,
                        )
                    )
                except Exception as e:
                    logger.warning(f'Failed to get live stream url: {repr(e)}')
                    observer.on_error(e)
                else:
                    logger.info(f"Got live stream url: '{url}'")
                    self._stream_url = url
                    self._stream_host = urlparse(url).hostname or ''
                    self._stream_params = params
                    observer.on_next(url)

            return source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe)

    def _can_resue_url(self, params: StreamParams) -> bool:
        if params == self._stream_params and self._stream_url:
            try:
                response = requests.get(
                    self._stream_url, stream=True, headers=self._live.headers
                )
                response.raise_for_status()
            except Exception:
                return False
            else:
                return True
        else:
            return False

    def _should_retry(self, exc: Exception) -> bool:
        if isinstance(
            exc,
            (
                NoStreamAvailable,
                NoStreamCodecAvailable,
                NoStreamFormatAvailable,
                NoStreamQualityAvailable,
                NoAlternativeStreamAvailable,
            ),
        ):
            return True
        else:
            return False

    def _before_retry(self, exc: Exception) -> None:
        try:
            raise exc
        except (NoStreamAvailable, NoStreamCodecAvailable, NoStreamFormatAvailable):
            pass
        except NoStreamQualityAvailable:
            qn = self._stream_param_holder.quality_number
            logger.info(
                f'The specified stream quality ({qn}) is not available, '
                'will using the original stream quality (10000) instead.'
            )
            self._stream_param_holder.fall_back_quality()
        except NoAlternativeStreamAvailable:
            logger.debug(
                'No alternative stream url available, '
                'will using the primary stream url instead.'
            )
            self._stream_param_holder.use_alternative_stream = False
            # self._stream_param_holder.rotate_api_platform()  # XXX: use web api only
        except LiveRoomHidden:
            pass
        except LiveRoomLocked:
            pass
        except LiveRoomEncrypted:
            pass
        except Exception:
            pass



from enum import Enum
from typing import Any, Dict, Literal, Optional

from pydantic import BaseModel


__all__ = (
    'ResponseMessage',
    'DataSelection',
    'AliasKeyOfSettings',
)


class ResponseMessage(BaseModel):
    code: int = 0
    message: str = ''
    data: Optional[Dict[str, Any]] = None


class DataSelection(str, Enum):
    ALL = 'all'

    # live status
    PREPARING = 'preparing'
    LIVING = 'living'
    ROUNDING = 'rounding'

    # task status
    MONITOR_ENABLED = 'monitor_enabled'
    MONITOR_DISABLED = 'monitor_disabled'
    RECORDER_ENABLED = 'recorder_enabled'
    RECORDER_DISABLED = 'recorder_disabled'

    # task running status
    STOPPED = 'stopped'
    WAITTING = 'waitting'
    RECORDING = 'recording'
    REMUXING = 'remuxing'
    INJECTING = 'injecting'


AliasKeyOfSettings = Literal[
    'version',
    'tasks',
    'output',
    'logging',
    'biliApi',
    'header',
    'danmaku',
    'recorder',
    'postprocessing',
    'space',
    'emailNotification',
    'serverchanNotification',
    'pushdeerNotification',
    'pushplusNotification',
    'telegramNotification',
    'barkNotification',
    'webhooks',
]



import asyncio
import html
import logging
from contextlib import suppress
from threading import Lock
from typing import Iterator, List, Optional

from tenacity import AsyncRetrying, retry_if_not_exception_type, stop_after_attempt

from blrec import __github__, __prog__, __version__
from blrec.bili.live import Live
from blrec.core.models import GiftSendMsg, GuardBuyMsg, SuperChatMsg, UserToastMsg
from blrec.danmaku.io import DanmakuWriter
from blrec.danmaku.models import (
    Danmu,
    GiftSendRecord,
    GuardBuyRecord,
    Metadata,
    SuperChatRecord,
    UserToast,
)
from blrec.event.event_emitter import EventEmitter, EventListener
from blrec.exception import exception_callback, submit_exception
from blrec.logging.room_id import aio_task_with_room_id
from blrec.path import danmaku_path
from blrec.utils.mixins import SwitchableMixin

from .danmaku_receiver import DanmakuReceiver, DanmuMsg
from .statistics import Statistics
from .stream_recorder import StreamRecorder, StreamRecorderEventListener

__all__ = 'DanmakuDumper', 'DanmakuDumperEventListener'


logger = logging.getLogger(__name__)


class DanmakuDumperEventListener(EventListener):
    async def on_danmaku_file_created(self, path: str) -> None:
        ...

    async def on_danmaku_file_completed(self, path: str) -> None:
        ...


class DanmakuDumper(
    EventEmitter[DanmakuDumperEventListener],
    StreamRecorderEventListener,
    SwitchableMixin,
):
    def __init__(
        self,
        live: Live,
        stream_recorder: StreamRecorder,
        danmaku_receiver: DanmakuReceiver,
        *,
        danmu_uname: bool = False,
        record_gift_send: bool = False,
        record_free_gifts: bool = False,
        record_guard_buy: bool = False,
        record_super_chat: bool = False,
    ) -> None:
        super().__init__()

        self._live = live
        self._stream_recorder = stream_recorder
        self._receiver = danmaku_receiver

        self.danmu_uname = danmu_uname
        self.record_gift_send = record_gift_send
        self.record_free_gifts = record_free_gifts
        self.record_guard_buy = record_guard_buy
        self.record_super_chat = record_super_chat

        self._lock: Lock = Lock()
        self._path: Optional[str] = None
        self._files: List[str] = []
        self._statistics = Statistics(interval=60)

    @property
    def danmu_total(self) -> int:
        return self._statistics.count

    @property
    def danmu_rate(self) -> float:
        return self._statistics.rate

    @property
    def elapsed(self) -> float:
        return self._statistics.elapsed

    @property
    def dumping_path(self) -> Optional[str]:
        return self._path

    def _do_enable(self) -> None:
        self._stream_recorder.add_listener(self)
        self._statistics.reset()
        logger.debug('Enabled danmaku dumper')

    def _do_disable(self) -> None:
        self._stream_recorder.remove_listener(self)
        asyncio.create_task(self._stop_dumping())
        self._statistics.freeze()
        logger.debug('Disabled danmaku dumper')

    def set_live_start_time(self, time: int) -> None:
        self._live_start_time = time

    def has_file(self) -> bool:
        return bool(self._files)

    def get_files(self) -> Iterator[str]:
        for file in self._files:
            yield file

    def clear_files(self) -> None:
        self._files.clear()

    async def on_video_file_created(
        self, video_path: str, record_start_time: int
    ) -> None:
        with self._lock:
            self._delta: float = 0
            self._record_start_time: int = record_start_time
            self._timebase: int = self._record_start_time * 1000
            self._stream_recording_interrupted: bool = False
            self._path = danmaku_path(video_path)
            self._files.append(self._path)
            await self._stop_dumping()
            self._start_dumping()

    async def on_video_file_completed(self, video_path: str) -> None:
        with self._lock:
            await self._stop_dumping()
            self._path = None

    async def on_stream_recording_interrupted(self, duration: float) -> None:
        logger.debug(f'Stream recording interrupted, {duration}')
        self._duration = duration
        self._stream_recording_recovered = asyncio.Condition()
        self._stream_recording_interrupted = True

    async def on_stream_recording_recovered(self, timestamp: int) -> None:
        logger.debug(f'Stream recording recovered, {timestamp}')
        self._timebase = timestamp * 1000
        self._delta = self._duration * 1000
        self._stream_recording_interrupted = False
        async with self._stream_recording_recovered:
            self._stream_recording_recovered.notify_all()

    def _start_dumping(self) -> None:
        self._create_dump_task()

    async def _stop_dumping(self) -> None:
        if hasattr(self, '_dump_task'):
            await self._cancel_dump_task()
            del self._dump_task  # type: ignore

    def _create_dump_task(self) -> None:
        self._dump_task = asyncio.create_task(self._do_dump())
        self._dump_task.add_done_callback(exception_callback)

    async def _cancel_dump_task(self) -> None:
        self._dump_task.cancel()
        with suppress(asyncio.CancelledError):
            await self._dump_task

    @aio_task_with_room_id
    async def _do_dump(self) -> None:
        assert self._path is not None
        logger.debug('Started dumping danmaku')

        try:
            async with DanmakuWriter(self._path) as writer:
                logger.info(f"Danmaku file created: '{self._path}'")
                await self._emit('danmaku_file_created', self._path)
                await writer.write_metadata(self._make_metadata())

                async for attempt in AsyncRetrying(
                    retry=retry_if_not_exception_type((asyncio.CancelledError)),
                    stop=stop_after_attempt(3),
                ):
                    with attempt:
                        try:
                            await self._dumping_loop(writer)
                        except Exception as e:
                            submit_exception(e)
                            raise
        finally:
            logger.info(f"Danmaku file completed: '{self._path}'")
            await self._emit('danmaku_file_completed', self._path)
            logger.debug('Stopped dumping danmaku')

    async def _dumping_loop(self, writer: DanmakuWriter) -> None:
        while True:
            msg = await self._receiver.get_message()

            if isinstance(msg, DanmuMsg):
                await writer.write_danmu(self._make_danmu(msg))
                self._statistics.submit(1)
            elif isinstance(msg, UserToastMsg):
                await writer.write_user_toast(self._make_user_toast(msg))
            elif isinstance(msg, GiftSendMsg):
                if not self.record_gift_send:
                    continue
                record = self._make_gift_send_record(msg)
                if not self.record_free_gifts and record.is_free_gift():
                    continue
                await writer.write_gift_send_record(record)
            elif isinstance(msg, GuardBuyMsg):
                if not self.record_guard_buy:
                    continue
                await writer.write_guard_buy_record(self._make_guard_buy_record(msg))
            elif isinstance(msg, SuperChatMsg):
                if not self.record_super_chat:
                    continue
                await writer.write_super_chat_record(self._make_super_chat_record(msg))
            else:
                logger.warning('Unsupported message type:', repr(msg))

            if self._stream_recording_interrupted:
                logger.debug(
                    f'Last message before stream recording interrupted: {repr(msg)}'
                )
                async with self._stream_recording_recovered:
                    await self._stream_recording_recovered.wait()

    def _make_metadata(self) -> Metadata:
        return Metadata(
            user_name=self._live.user_info.name,
            room_id=self._live.room_info.room_id,
            room_title=self._live.room_info.title,
            area=self._live.room_info.area_name,
            parent_area=self._live.room_info.parent_area_name,
            live_start_time=self._live.room_info.live_start_time,
            record_start_time=self._record_start_time,
            recorder=f'{__prog__} v{__version__} {__github__}',
        )

    def _make_danmu(self, msg: DanmuMsg) -> Danmu:
        if self.danmu_uname:
            text = f'{msg.uname}: {msg.text}'
        else:
            text = msg.text
        text = html.escape(text)

        return Danmu(
            stime=self._calc_stime(msg.date),
            mode=msg.mode,
            size=msg.size,
            color=msg.color,
            date=msg.date,
            pool=msg.pool,
            uid_hash=msg.uid_hash,
            uid=msg.uid,
            uname=msg.uname,
            dmid=msg.dmid,
            text=text,
        )

    def _make_gift_send_record(self, msg: GiftSendMsg) -> GiftSendRecord:
        return GiftSendRecord(
            ts=self._calc_stime(msg.timestamp * 1000),
            uid=msg.uid,
            user=msg.uname,
            giftname=msg.gift_name,
            giftcount=msg.count,
            cointype=msg.coin_type,
            price=msg.price,
        )

    def _make_guard_buy_record(self, msg: GuardBuyMsg) -> GuardBuyRecord:
        return GuardBuyRecord(
            ts=self._calc_stime(msg.timestamp * 1000),
            uid=msg.uid,
            user=msg.uname,
            giftname=msg.gift_name,
            count=msg.count,
            price=msg.price,
            level=msg.guard_level,
        )

    def _make_super_chat_record(self, msg: SuperChatMsg) -> SuperChatRecord:
        return SuperChatRecord(
            ts=self._calc_stime(msg.timestamp * 1000),
            uid=msg.uid,
            user=msg.uname,
            price=msg.price * msg.rate,
            time=msg.time,
            message=msg.message,
        )

    def _make_user_toast(self, msg: UserToastMsg) -> UserToast:
        return UserToast(
            ts=self._calc_stime(msg.start_time * 1000),
            uid=msg.uid,
            user=msg.username,
            unit=msg.unit,
            count=msg.num,
            price=msg.price,
            role=msg.role_name,
            level=msg.guard_level,
            msg=msg.toast_msg,
        )

    def _calc_stime(self, timestamp: int) -> float:
        return (max(timestamp - self._timebase, 0) + self._delta) / 1000



from .notifiers import (
    Notifier,
    MessageNotifier,
    EmailNotifier,
    ServerchanNotifier,
    PushdeerNotifier,
    PushplusNotifier,
    TelegramNotifier,
    BarkNotifier,
)
from .providers import (
    MessagingProvider,
    EmailService,
    Serverchan,
    Pushdeer,
    Pushplus,
    Telegram,
    Bark,
)


__all__ = (
    'MessagingProvider',
    'EmailService',
    'Serverchan',
    'Pushdeer',
    'Pushplus',
    'Telegram',
    'Bark',

    'Notifier',
    'MessageNotifier',
    'EmailNotifier',
    'ServerchanNotifier',
    'PushdeerNotifier',
    'PushplusNotifier',
    'TelegramNotifier',
    'BarkNotifier',
)



from urllib.parse import urlparse, urlunparse
from typing import Literal


def ensure_scheme(url: str, scheme: Literal['http', 'https']) -> str:
    return urlunparse(urlparse(url)._replace(scheme=scheme))



from typing import Optional

from fastapi import APIRouter, Depends

from ..responses import not_found_responses
from ..dependencies import settings_include_set, settings_exclude_set
from ...setting import (
    SettingsIn,
    SettingsOut,
    TaskOptions,
)
from ...setting.typing import KeySetOfSettings
from ...application import Application


app: Application = None  # type: ignore  # bypass flake8 F821

router = APIRouter(
    prefix='/api/v1/settings',
    tags=['settings'],
)


@router.get(
    '',
    response_model=SettingsOut,
    response_model_exclude_unset=True,
)
async def get_settings(
    include: Optional[KeySetOfSettings] = Depends(settings_include_set),
    exclude: Optional[KeySetOfSettings] = Depends(settings_exclude_set),
) -> SettingsOut:
    return app.get_settings(include, exclude)


@router.patch(
    '',
    response_model=SettingsOut,
    response_model_exclude_unset=True,
)
async def change_settings(settings: SettingsIn) -> SettingsOut:
    """Change settings of the application

    Change network request headers will cause
    **all** the Danmaku client be **reconnected**!
    """
    return await app.change_settings(settings)


@router.get(
    '/tasks/{room_id}',
    response_model=TaskOptions,
    response_model_exclude_unset=True,
    responses={**not_found_responses},
)
async def get_task_options(room_id: int) -> TaskOptions:
    return app.get_task_options(room_id)


@router.patch(
    '/tasks/{room_id}',
    response_model=TaskOptions,
    responses={**not_found_responses},
)
async def change_task_options(
    room_id: int, options: TaskOptions
) -> TaskOptions:
    """Change task-specific options

    Task-specific options will shadow the corresponding global settings.
    Explicitly set options to **null** will remove the value shadowing.

    Change network request headers will cause
    the Danmaku client be **reconnected**!
    """
    return await app.change_task_options(room_id, options)



import os
import glob
from datetime import datetime
from pathlib import Path
import logging
import asyncio
from functools import partial
from typing import Iterable, List

from tenacity import retry, wait_none, stop_after_attempt, retry_if_exception_type

from .helpers import delete_file, is_space_enough
from .space_monitor import SpaceMonitor, DiskUsage, SpaceEventListener
from ..utils.mixins import SwitchableMixin


__all__ = ('SpaceReclaimer',)


logger = logging.getLogger(__name__)


class SpaceReclaimer(SpaceEventListener, SwitchableMixin):
    _SUFFIX_SET = frozenset(
        ('.flv', '.mp4', '.ts', '.m4s', '.m3u8' '.xml', '.jsonl', '.jpg')
    )

    def __init__(
        self,
        space_monitor: SpaceMonitor,
        path: str,
        *,
        rec_ttl: int = 60 * 60 * 24,
        recycle_records: bool = False,
    ) -> None:
        super().__init__()
        self._space_monitor = space_monitor
        self.path = path
        if value := os.environ.get('BLREC_REC_TTL'):
            try:
                rec_ttl = int(value)
            except Exception as exc:
                logger.warning(repr(exc))
        self.rec_ttl = rec_ttl
        self.recycle_records = recycle_records

    async def on_space_no_enough(
        self, path: str, threshold: int, disk_usage: DiskUsage
    ) -> None:
        await self._free_space(threshold)

    def _do_enable(self) -> None:
        self._space_monitor.add_listener(self)
        logger.debug('Enabled space reclaimer')

    def _do_disable(self) -> None:
        self._space_monitor.remove_listener(self)
        logger.debug('Disabled space reclaimer')

    async def _free_space(self, size: int) -> bool:
        if is_space_enough(self.path, size):
            return True
        if self.recycle_records:
            if await self._free_space_from_records(size):
                return True
        return False

    async def _free_space_from_records(self, size: int) -> bool:
        logger.info('Free space from records ...')
        ts = datetime.now().timestamp() - self.rec_ttl
        for path in await self._get_record_file_paths(ts):
            await delete_file(path)
            if is_space_enough(self.path, size):
                return True
        return False

    @retry(
        retry=retry_if_exception_type(OSError),
        wait=wait_none(),
        stop=stop_after_attempt(3),
    )
    async def _get_record_file_paths(self, ts: float) -> List[str]:
        glob_path = os.path.join(self.path, '*/**/*.*')
        paths: Iterable[Path]
        paths = map(lambda p: Path(p), glob.iglob(glob_path, recursive=True))
        paths = filter(lambda p: p.suffix in self._SUFFIX_SET, paths)
        paths = filter(lambda p: p.stat().st_mtime < ts > p.stat().st_atime, paths)
        func = partial(
            sorted, paths, key=lambda p: (p.stat().st_mtime, p.stat().st_atime)
        )
        loop = asyncio.get_running_loop()
        path_list = await loop.run_in_executor(None, func)
        return list(map(str, path_list))



import asyncio
import json
import logging
import os
import shutil
from pathlib import PurePath
from typing import Any, Dict, Iterable, Literal

import aiofiles

from blrec.path.helpers import (
    cover_path,
    danmaku_path,
    raw_danmaku_path,
    record_metadata_path,
)

from ..flv.helpers import get_extra_metadata as _get_extra_metadata
from ..flv.helpers import get_metadata as _get_metadata

logger = logging.getLogger(__name__)


async def discard_files(
    paths: Iterable[str], log_level: Literal['INFO', 'DEBUG'] = 'INFO'
) -> None:
    for path in paths:
        await discard_file(path, log_level)


async def discard_file(path: str, log_level: Literal['INFO', 'DEBUG'] = 'INFO') -> None:
    from ..disk_space import delete_file

    await delete_file(path, log_level)


async def discard_dir(path: str, log_level: Literal['INFO', 'DEBUG'] = 'INFO') -> None:
    loop = asyncio.get_running_loop()
    try:
        await loop.run_in_executor(None, shutil.rmtree, path)
    except Exception as e:
        logger.error(f'Failed to delete {path!r}, due to: {repr(e)}')
    else:
        logger.log(logging.getLevelName(log_level), f'Deleted {path!r}')


def files_related(video_path: str) -> Iterable[str]:
    for path in [
        danmaku_path(video_path),
        raw_danmaku_path(video_path),
        cover_path(video_path, ext='jpg'),
        cover_path(video_path, ext='png'),
    ]:
        if os.path.isfile(path):
            yield path


async def copy_files_related(video_path: str) -> None:
    loop = asyncio.get_running_loop()
    dirname = os.path.dirname(video_path)

    for src_path in files_related(video_path):
        root, ext = os.path.splitext(src_path)
        dst_path = PurePath(dirname).with_suffix(ext)
        try:
            await loop.run_in_executor(None, shutil.copy, src_path, dst_path)
        except Exception as e:
            logger.error(f"Failed to copy '{src_path}' to '{dst_path}': {repr(e)}")
        else:
            logger.info(f"Copied '{src_path}' to '{dst_path}'")


async def get_metadata(flv_path: str) -> Dict[str, Any]:
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, _get_metadata, flv_path)


async def get_record_metadata(video_path: str) -> Dict[str, Any]:
    if video_path.endswith('.m3u8'):
        path = record_metadata_path(video_path)
    else:
        raise NotImplementedError(video_path)
    async with aiofiles.open(path, 'rb') as file:
        data = await file.read()
        return json.loads(data)


async def get_extra_metadata(flv_path: str) -> Dict[str, Any]:
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, _get_extra_metadata, flv_path)



from typing import Literal

import attr


@attr.s(auto_attribs=True, slots=True, frozen=True)
class Metadata:
    user_name: str
    room_id: int
    room_title: str
    area: str
    parent_area: str
    live_start_time: int  # seconds
    record_start_time: int  # seconds
    recorder: str


@attr.s(auto_attribs=True, slots=True, frozen=True)
class Danmu:
    stime: float
    mode: int
    size: int
    color: int
    date: int  # milliseconds
    pool: int
    uid_hash: str
    uid: int
    uname: str
    dmid: int
    text: str


@attr.s(auto_attribs=True, slots=True, frozen=True)
class UserToast:
    ts: float  # start_time
    uid: int
    user: str  # username
    unit: str
    count: int  # num
    price: int
    role: str  # role_name
    level: str  # guard_level
    msg: str  # toast_msg


@attr.s(auto_attribs=True, slots=True, frozen=True)
class GiftSendRecord:
    ts: float
    uid: int
    user: str
    giftname: str
    giftcount: int
    cointype: Literal['sliver', 'gold']
    price: int

    def is_free_gift(self) -> bool:
        return self.cointype != 'gold'


@attr.s(auto_attribs=True, slots=True, frozen=True)
class GuardBuyRecord:
    ts: float
    uid: int
    user: str
    giftname: str
    count: int
    price: int
    level: int


@attr.s(auto_attribs=True, slots=True, frozen=True)
class SuperChatRecord:
    ts: float
    uid: int
    user: str
    price: int
    time: int
    message: str



import logging
import secrets
from typing import Dict, Optional, Set

from fastapi import Header, Request, status
from fastapi.exceptions import HTTPException

logger = logging.getLogger(__name__)


api_key = ''

MAX_WHITELIST = 100
MAX_BLACKLIST = 100
MAX_ATTEMPTING_CLIENTS = 100
MAX_ATTEMPTS = 3
whitelist: Set[str] = set()
blacklist: Set[str] = set()
attempting_clients: Dict[str, int] = {}


async def authenticate(
    request: Request, x_api_key: Optional[str] = Header(None)
) -> None:
    assert api_key, 'api_key is required'

    if not x_api_key:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED, detail='No api key'
        )

    assert request.client is not None, 'client should not be None'
    client_ip = request.client.host
    assert client_ip, 'client_ip is required'

    if client_ip in blacklist:
        raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail='Blacklisted')
    if client_ip not in whitelist:
        if len(whitelist) >= MAX_WHITELIST or len(blacklist) >= MAX_BLACKLIST:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail='Max clients allowed in whitelist or blacklist ' 'will exceeded',
            )
        if len(attempting_clients) >= MAX_ATTEMPTING_CLIENTS:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail='Max attempting clients allowed exceeded',
            )
        if client_ip not in attempting_clients:
            attempting_clients[client_ip] = 1
        else:
            attempting_clients[client_ip] += 1
        if attempting_clients[client_ip] > MAX_ATTEMPTS:
            del attempting_clients[client_ip]
            blacklist.add(client_ip)
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail='Max api key attempts exceeded',
            )

    if not secrets.compare_digest(x_api_key, api_key):
        if client_ip in whitelist:
            whitelist.remove(client_ip)
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED, detail='API key is invalid'
        )

    if client_ip in attempting_clients:
        del attempting_clients[client_ip]
    if client_ip not in whitelist:
        whitelist.add(client_ip)



import logging
from typing import Iterable


from .io import DanmakuReader, DanmakuWriter
from .common import copy_damus
from .typing import TimebaseType


__all__ = 'TimebaseType', 'DanmakuCombinator'


logger = logging.getLogger(__name__)


class DanmakuCombinator:
    def __init__(
        self,
        in_paths: Iterable[str],
        out_path: str,
        timebase_type: TimebaseType = TimebaseType.RECORD,
    ) -> None:
        self._in_paths = list(in_paths)
        self._out_path = out_path
        self._timebase_type = timebase_type

    async def combine(self) -> None:
        async with DanmakuWriter(self._out_path) as writer:
            writer = writer

            async with DanmakuReader(self._in_paths[0]) as reader:
                metadata = await reader.read_metadata()
                await writer.write_metadata(metadata)

                if self._timebase_type == TimebaseType.LIVE:
                    timebase = metadata.live_start_time * 1000
                else:
                    timebase = metadata.record_start_time * 1000

                await copy_damus(reader, writer, timebase=timebase)

            for path in self._in_paths[1:]:
                async with DanmakuReader(path) as reader:
                    await copy_damus(reader, writer, timebase=timebase)



import logging
from typing import Callable

from reactivex import operators as ops

from ..common import is_avc_end_sequence_tag
from .concat import concat
from .correct import correct
from .defragment import defragment
from .fix import fix
from .sort import sort
from .split import split
from .typing import FLVStream

__all__ = ('process',)

logger = logging.getLogger(__name__)


def process(sort_tags: bool = False) -> Callable[[FLVStream], FLVStream]:
    def _process(source: FLVStream) -> FLVStream:
        if sort_tags:
            return source.pipe(
                defragment(),
                split(),
                sort(),
                ops.filter(lambda v: not is_avc_end_sequence_tag(v)),  # type: ignore
                correct(),
                fix(),
                concat(),
            )
        else:
            return source.pipe(
                defragment(),
                split(),
                ops.filter(lambda v: not is_avc_end_sequence_tag(v)),  # type: ignore
                correct(),
                fix(),
                concat(),
            )

    return _process



from .playlist_dumper import PlaylistDumper
from .playlist_fetcher import PlaylistFetcher
from .playlist_resolver import PlaylistResolver
from .prober import Prober, StreamProfile
from .segment_dumper import SegmentDumper
from .segment_fetcher import InitSectionData, SegmentData, SegmentFetcher
from .segment_parser import SegmentParser
from .segment_remuxer import SegmentRemuxer

__all__ = (
    'InitSectionData',
    'PlaylistDumper',
    'PlaylistFetcher',
    'PlaylistResolver',
    'Prober',
    'SegmentData',
    'SegmentDumper',
    'SegmentFetcher',
    'SegmentParser',
    'SegmentRemuxer',
    'StreamProfile',
)



import attr
import aiohttp


@attr.s(auto_attribs=True, frozen=True, slots=True)
class ApiRequestError(Exception):
    code: int
    message: str


class DanmakuClientAuthError(aiohttp.ClientError):
    pass


class LiveRoomHidden(Exception):
    pass


class LiveRoomLocked(Exception):
    pass


class LiveRoomEncrypted(Exception):
    pass


class NoStreamAvailable(Exception):
    pass


class NoStreamFormatAvailable(Exception):
    pass


class NoStreamCodecAvailable(Exception):
    pass


class NoStreamQualityAvailable(Exception):
    pass


class NoAlternativeStreamAvailable(Exception):
    pass



from .task_manager import RecordTaskManager
from .models import (
    TaskData,
    TaskStatus,
    TaskParam,
    RunningStatus,
    VideoFileDetail,
    DanmakuFileDetail,
)


__all__ = (
    'RecordTaskManager',

    'TaskData',
    'TaskStatus',
    'TaskParam',
    'RunningStatus',
    'VideoFileDetail',
    'DanmakuFileDetail',
)



import asyncio
import logging
import os
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Final, Optional, Tuple

import attr
import humanize
from liquid import Environment
from liquid.filter import math_filter
from pkg_resources import resource_string
from tenacity import (
    AsyncRetrying,
    retry_if_exception,
    stop_after_delay,
    wait_exponential,
)

from ..event import (
    Error,
    ErrorData,
    EventCenter,
    LiveBeganEvent,
    LiveEndedEvent,
    SpaceNoEnoughEvent,
)
from ..event.typing import Event
from ..exception import ExceptionCenter, format_exception
from ..setting.typing import MessageType
from ..utils.mixins import SwitchableMixin
from .providers import (
    EmailService,
    MessagingProvider,
    Pushdeer,
    Pushplus,
    Serverchan,
    Telegram,
    Bark,
)

__all__ = (
    'Notifier',
    'MessageNotifier',
    'EmailNotifier',
    'ServerchanNotifier',
    'PushdeerNotifier',
    'PushplusNotifier',
    'TelegramNotifier',
    'BarkNotifer',
)


logger = logging.getLogger(__name__)


class Notifier(SwitchableMixin, ABC):
    def __init__(
        self,
        *,
        notify_began: bool = False,
        notify_ended: bool = False,
        notify_space: bool = False,
        notify_error: bool = False,
        began_message_type: MessageType = 'text',
        began_message_title: str = '',
        began_message_content: str = '',
        ended_message_type: MessageType = 'text',
        ended_message_title: str = '',
        ended_message_content: str = '',
        space_message_type: MessageType = 'text',
        space_message_title: str = '',
        space_message_content: str = '',
        error_message_type: MessageType = 'text',
        error_message_title: str = '',
        error_message_content: str = '',
    ) -> None:
        super().__init__()

        self._liquid_env = Environment()
        self._liquid_env.add_filter('intcomma', math_filter(humanize.intcomma))
        self._liquid_env.add_filter('naturalsize', math_filter(humanize.naturalsize))
        self._liquid_env.add_filter(
            'datetimestring',
            math_filter(lambda n: datetime.fromtimestamp(n).isoformat()),
        )

        self.notify_began = notify_began
        self.notify_ended = notify_ended
        self.notify_space = notify_space
        self.notify_error = notify_error
        self.began_message_type = began_message_type
        self.began_message_title = began_message_title
        self.began_message_content = began_message_content
        self.ended_message_type = ended_message_type
        self.ended_message_title = ended_message_title
        self.ended_message_content = ended_message_content
        self.space_message_type = space_message_type
        self.space_message_title = space_message_title
        self.space_message_content = space_message_content
        self.error_message_type = error_message_type
        self.error_message_title = error_message_title
        self.error_message_content = error_message_content

    def _do_enable(self) -> None:
        events = EventCenter.get_instance().events
        self._event_subscription = events.subscribe(self._on_event)
        exceptions = ExceptionCenter.get_instance().exceptions
        self._exc_subscription = exceptions.subscribe(self._on_exception)

    def _do_disable(self) -> None:
        self._event_subscription.dispose()
        self._exc_subscription.dispose()

    def _on_event(self, event: Event) -> None:
        if isinstance(event, LiveBeganEvent):
            if self._should_notify_live_began(event):
                self._notify_live_began(event)
        elif isinstance(event, LiveEndedEvent):
            if self._should_notify_live_ended(event):
                self._notify_live_ended(event)
        elif isinstance(event, SpaceNoEnoughEvent):
            if self._should_notify_space_no_enough(event):
                self._notify_space_no_enough(event)
        else:
            pass

    def _on_exception(self, exc: BaseException) -> None:
        if self._should_notify_exception(exc):
            error = Error.from_data(ErrorData.from_exc(exc))
            self._notify_exception(error)

    def _should_notify_live_began(self, event: LiveBeganEvent) -> bool:
        return self.notify_began

    def _should_notify_live_ended(self, event: LiveEndedEvent) -> bool:
        return self.notify_ended

    def _should_notify_space_no_enough(self, event: SpaceNoEnoughEvent) -> bool:
        return self.notify_space

    def _should_notify_exception(self, exc: BaseException) -> bool:
        return self.notify_error

    @abstractmethod
    def _notify_live_began(self, event: LiveBeganEvent) -> None:
        raise NotImplementedError

    @abstractmethod
    def _notify_live_ended(self, event: LiveEndedEvent) -> None:
        raise NotImplementedError

    @abstractmethod
    def _notify_space_no_enough(self, event: SpaceNoEnoughEvent) -> None:
        raise NotImplementedError

    @abstractmethod
    def _notify_exception(self, event: Error) -> None:
        raise NotImplementedError

    @abstractmethod
    def _make_began_message(self, event: LiveBeganEvent) -> Tuple[str, str]:
        raise NotImplementedError

    @abstractmethod
    def _make_ended_message(self, event: LiveEndedEvent) -> Tuple[str, str]:
        raise NotImplementedError

    @abstractmethod
    def _make_space_message(self, event: SpaceNoEnoughEvent) -> Tuple[str, str]:
        raise NotImplementedError

    @abstractmethod
    def _make_error_message(self, event: Error) -> Tuple[str, str]:
        raise NotImplementedError


BEGAN_MESSAGE_TITLE: Final[str] = '{{ event.data.user_info.name }} 开播啦'
ENDED_MESSAGE_TITLE: Final[str] = '{{ event.data.user_info.name }} 下播了'
SPACE_MESSAGE_TITLE: Final[str] = '空间不足！'
ERROR_MESSAGE_TITLE: Final[str] = '出错了~'


class MessageNotifier(Notifier, ABC):
    provider: MessagingProvider

    def _notify_live_began(self, event: LiveBeganEvent) -> None:
        title, content = self._make_began_message(event)
        self._send_message(title, content, self.began_message_type)

    def _notify_live_ended(self, event: LiveEndedEvent) -> None:
        title, content = self._make_ended_message(event)
        self._send_message(title, content, self.ended_message_type)

    def _notify_space_no_enough(self, event: SpaceNoEnoughEvent) -> None:
        title, content = self._make_space_message(event)
        self._send_message(title, content, self.space_message_type)

    def _notify_exception(self, event: Error) -> None:
        title, content = self._make_error_message(event)
        self._send_message(title, content, self.error_message_type)

    def _make_began_message(self, event: LiveBeganEvent) -> Tuple[str, str]:
        env = os.environ.copy()
        try:
            template = self._liquid_env.from_string(self._get_began_message_title())
            title = template.render(event=attr.asdict(event), env=env)
            template = self._liquid_env.from_string(self._get_began_message_content())
            content = template.render(event=attr.asdict(event), env=env)
        except Exception as e:
            logger.warning(f'Failed to render began message template: {repr(e)}')
            title = f'{event.data.user_info.name} 开播啦'
            content = format_exception(e)
        return title, content

    def _make_ended_message(self, event: LiveEndedEvent) -> Tuple[str, str]:
        env = os.environ.copy()
        try:
            template = self._liquid_env.from_string(self._get_ended_message_title())
            title = template.render(event=attr.asdict(event), env=env)
            template = self._liquid_env.from_string(self._get_ended_message_content())
            content = template.render(event=attr.asdict(event), env=env)
        except Exception as e:
            logger.warning(f'Failed to render ended message template: {repr(e)}')
            title = f'{event.data.user_info.name} 下播了'
            content = format_exception(e)
        return title, content

    def _make_space_message(self, event: SpaceNoEnoughEvent) -> Tuple[str, str]:
        env = os.environ.copy()
        try:
            template = self._liquid_env.from_string(self._get_space_message_title())
            title = template.render(event=attr.asdict(event), env=env)
            template = self._liquid_env.from_string(self._get_space_message_content())
            content = template.render(event=attr.asdict(event), env=env)
        except Exception as e:
            logger.warning(f'Failed to render space message template: {repr(e)}')
            title = '空间不足！'
            content = format_exception(e)
        return title, content

    def _make_error_message(self, event: Error) -> Tuple[str, str]:
        env = os.environ.copy()
        try:
            template = self._liquid_env.from_string(self._get_error_message_title())
            title = template.render(event=attr.asdict(event), env=env)
            template = self._liquid_env.from_string(self._get_error_message_content())
            content = template.render(event=attr.asdict(event), env=env)
        except Exception as e:
            logger.warning(f'Failed to render error message template: {repr(e)}')
            title = '出错了~'
            content = format_exception(e)
        return title, content

    def _send_message(self, title: str, content: str, msg_type: MessageType) -> None:
        asyncio.create_task(self._send_message_async(title, content, msg_type))

    async def _send_message_async(
        self, title: str, content: str, msg_type: MessageType
    ) -> None:
        try:
            async for attempt in AsyncRetrying(
                reraise=True,
                stop=stop_after_delay(300),
                wait=wait_exponential(multiplier=0.1, max=10),
                retry=retry_if_exception(lambda e: not isinstance(e, ValueError)),
            ):
                with attempt:
                    await self.provider.send_message(title, content, msg_type)
        except Exception as e:
            logger.warning(
                'Failed to send a message via {}: {}'.format(
                    self.provider.__class__.__name__, repr(e)
                )
            )

    def _get_began_message_title(self) -> str:
        return self.began_message_title or BEGAN_MESSAGE_TITLE

    def _get_began_message_content(self, msg_type: Optional[MessageType] = None) -> str:
        if self.began_message_content:
            return self.began_message_content
        msg_type = msg_type or self.began_message_type
        if msg_type == 'markdown':
            relpath = '../data/message_templates/markdown/live-began.md'
        elif msg_type == 'html':
            relpath = '../data/message_templates/html/live-began.html'
        else:
            relpath = '../data/message_templates/text/live-began.txt'
        return resource_string(__name__, relpath).decode('utf-8')

    def _get_ended_message_title(self) -> str:
        return self.ended_message_title or ENDED_MESSAGE_TITLE

    def _get_ended_message_content(self, msg_type: Optional[MessageType] = None) -> str:
        if self.ended_message_content:
            return self.ended_message_content
        msg_type = msg_type or self.ended_message_type
        if msg_type == 'markdown':
            relpath = '../data/message_templates/markdown/live-ended.md'
        elif msg_type == 'html':
            relpath = '../data/message_templates/html/live-ended.html'
        else:
            relpath = '../data/message_templates/text/live-ended.txt'
        return resource_string(__name__, relpath).decode('utf-8')

    def _get_space_message_title(self) -> str:
        return self.space_message_title or SPACE_MESSAGE_TITLE

    def _get_space_message_content(self, msg_type: Optional[MessageType] = None) -> str:
        if self.space_message_content:
            return self.space_message_content
        msg_type = msg_type or self.space_message_type
        if msg_type == 'markdown':
            relpath = '../data/message_templates/markdown/space-no-enough.md'
        elif msg_type == 'html':
            relpath = '../data/message_templates/html/space-no-enough.html'
        else:
            relpath = '../data/message_templates/text/space-no-enough.txt'
        return resource_string(__name__, relpath).decode('utf-8')

    def _get_error_message_title(self) -> str:
        return self.error_message_title or ERROR_MESSAGE_TITLE

    def _get_error_message_content(self, msg_type: Optional[MessageType] = None) -> str:
        if self.error_message_content:
            return self.error_message_content
        msg_type = msg_type or self.error_message_type
        if msg_type == 'markdown':
            relpath = '../data/message_templates/markdown/error.md'
        elif msg_type == 'html':
            relpath = '../data/message_templates/html/error.html'
        else:
            relpath = '../data/message_templates/text/error.txt'
        return resource_string(__name__, relpath).decode('utf-8')


class EmailNotifier(MessageNotifier):
    provider = EmailService.get_instance()

    def _do_enable(self) -> None:
        super()._do_enable()
        logger.debug('Enabled Email notifier')

    def _do_disable(self) -> None:
        super()._do_disable()
        logger.debug('Disabled Email notifier')


class ServerchanNotifier(MessageNotifier):
    provider = Serverchan.get_instance()

    def _do_enable(self) -> None:
        super()._do_enable()
        logger.debug('Enabled Serverchan notifier')

    def _do_disable(self) -> None:
        super()._do_disable()
        logger.debug('Disabled Serverchan notifier')


class PushdeerNotifier(MessageNotifier):
    provider = Pushdeer.get_instance()

    def _do_enable(self) -> None:
        super()._do_enable()
        logger.debug('Enabled Pushdeer notifier')

    def _do_disable(self) -> None:
        super()._do_disable()
        logger.debug('Disabled Pushdeer notifier')


class PushplusNotifier(MessageNotifier):
    provider = Pushplus.get_instance()

    def _do_enable(self) -> None:
        super()._do_enable()
        logger.debug('Enabled Pushplus notifier')

    def _do_disable(self) -> None:
        super()._do_disable()
        logger.debug('Disabled Pushplus notifier')


class TelegramNotifier(MessageNotifier):
    provider = Telegram.get_instance()

    def _do_enable(self) -> None:
        super()._do_enable()
        logger.debug('Enabled Telegram notifier')

    def _do_disable(self) -> None:
        super()._do_disable()
        logger.debug('Disabled Telegram notifier')

    def _get_began_message_content(self, msg_type: Optional[MessageType] = None) -> str:
        return super()._get_began_message_content(msg_type='text')

    def _get_ended_message_content(self, msg_type: Optional[MessageType] = None) -> str:
        return super()._get_ended_message_content(msg_type='text')

    def _get_space_message_content(self, msg_type: Optional[MessageType] = None) -> str:
        return super()._get_space_message_content(msg_type='text')

    def _get_error_message_content(self, msg_type: Optional[MessageType] = None) -> str:
        return super()._get_error_message_content(msg_type='text')


class BarkNotifier(MessageNotifier):
    provider = Bark.get_instance()

    def _do_enable(self) -> None:
        super()._do_enable()
        logger.debug('Enabled Bark notifier')

    def _do_disable(self) -> None:
        super()._do_disable()
        logger.debug('Disabled Bark notifier')



import logging
import os
from datetime import datetime
from typing import Any, Dict

import attr
from reactivex import Observable
from reactivex import operators as ops

from ..utils import operators as utils_ops
from . import operators as flv_ops
from .helpers import make_comment_for_joinpoints
from .operators import JoinPoint
from .operators.helpers import from_file

__all__ = 'InjectingProgress', 'inject_metadata'


logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class InjectingProgress:
    count: int
    total: int


def inject_metadata(
    path: str, metadata: Dict[str, Any], *, display_progress: bool = False
) -> Observable[InjectingProgress]:
    filesize = os.path.getsize(path)

    root, ext = os.path.splitext(path)
    temp_path = f'{root}_injecting{ext}'
    filename = os.path.basename(path)

    def metadata_provider(original_metadata: Dict[str, Any]) -> Dict[str, Any]:
        if joinpoints := metadata.get('joinpoints'):
            join_points = map(JoinPoint.from_metadata_value, joinpoints)
            if comment := original_metadata.get('Comment'):
                comment += '\n' + make_comment_for_joinpoints(join_points)
            else:
                comment = make_comment_for_joinpoints(join_points)
            metadata['Comment'] = comment
        return metadata

    return from_file(path).pipe(
        flv_ops.Injector(metadata_provider),
        flv_ops.Dumper(lambda: (temp_path, int(datetime.now().timestamp()))),
        flv_ops.ProgressBar(
            desc='Injecting',
            postfix=filename,
            total=filesize,
            disable=not display_progress,
        ),
        ops.map(lambda i: len(i)),  # type: ignore
        ops.scan(lambda acc, x: acc + x, 0),  # type: ignore
        ops.map(lambda s: InjectingProgress(s, filesize)),  # type: ignore
        utils_ops.replace(temp_path, path),
    )



import asyncio
import atexit
import logging
import os
import threading
from datetime import datetime
from logging import Handler, LogRecord
from logging.handlers import TimedRotatingFileHandler
from typing import Any, List, Optional

from colorama import Back, Fore, Style, deinit, init
from tqdm import tqdm

from .typing import LOG_LEVEL

__all__ = 'configure_logger', 'ConsoleHandler', 'TqdmOutputStream'


DISPLAY_PROGRESS = bool(os.environ.get('BLREC_PROGRESS'))


class TqdmOutputStream:
    def write(self, string: str = '') -> None:
        tqdm.write(string, end='')


class ConsoleHandler(logging.StreamHandler):
    def __init__(self, stream=None) -> None:  # type: ignore
        super().__init__(stream)

    def format(self, record: LogRecord) -> str:
        msg = super().format(record)

        level = record.levelno
        if level == logging.DEBUG:
            style = Fore.GREEN
        elif level == logging.WARNING:
            style = Fore.YELLOW
        elif level == logging.ERROR:
            style = Fore.RED
        elif level == logging.CRITICAL:
            style = Fore.WHITE + Back.RED + Style.BRIGHT
        else:
            style = ''

        return style + msg + Style.RESET_ALL if style else msg


_old_factory = logging.getLogRecordFactory()


def obtain_room_id() -> str:
    try:
        task = asyncio.current_task()
        assert task is not None
    except Exception:
        name = threading.current_thread().getName()
    else:
        name = task.get_name()

    if '::' in name:
        if room_id := name.split('::')[-1]:
            return room_id

    return ''


def record_factory(*args: Any, **kwargs: Any) -> LogRecord:
    record = _old_factory(*args, **kwargs)

    if room_id := obtain_room_id():
        record.roomid = '[' + room_id + '] '  # type: ignore
    else:
        record.roomid = ''  # type: ignore

    return record


logging.setLogRecordFactory(record_factory)


_old_handlers: List[Handler] = []


def configure_logger(
    log_dir: str,
    *,
    console_log_level: LOG_LEVEL = 'INFO',
    backup_count: Optional[int] = None,
) -> None:
    # config root logger
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    # config formatter
    formatter = logging.Formatter(
        '[%(asctime)s] [%(levelname)s] [%(module)s] %(roomid)s%(message)s'
    )

    # logging to console
    if DISPLAY_PROGRESS:
        console_handler = ConsoleHandler(TqdmOutputStream())
    else:
        console_handler = ConsoleHandler()
    console_handler.setLevel(logging.getLevelName(console_log_level))
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    # logging to file
    log_file_path = make_log_file_path(log_dir)
    file_handler = TimedRotatingFileHandler(
        filename=log_file_path,
        when='MIDNIGHT',
        backupCount=backup_count or 0,
        encoding='utf-8',
    )
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    # remove old handlers after re-configured
    for handler in _old_handlers:
        logger.removeHandler(handler)

    # retain old handlers for the removing
    _old_handlers.append(console_handler)
    _old_handlers.append(file_handler)

    logger.info(f'log file: {log_file_path}')


def make_log_file_path(log_dir: str) -> str:
    data_time_string = datetime.now().strftime("%Y-%m-%d-%H%M%S-%f")
    filename = f'blrec_{data_time_string}.log'
    return os.path.abspath(os.path.join(log_dir, filename))


init()
atexit.register(deinit)



import logging
from typing import Optional

from reactivex.scheduler import NewThreadScheduler

from blrec.bili.live import Live
from blrec.bili.typing import QualityNumber
from blrec.flv import operators as flv_ops
from blrec.flv.metadata_dumper import MetadataDumper
from blrec.utils.mixins import SupportDebugMixin

from . import operators as core_ops
from .stream_recorder_impl import StreamRecorderImpl

__all__ = ('FLVStreamRecorderImpl',)


logger = logging.getLogger(__name__)


class FLVStreamRecorderImpl(StreamRecorderImpl, SupportDebugMixin):
    def __init__(
        self,
        live: Live,
        out_dir: str,
        path_template: str,
        *,
        quality_number: QualityNumber = 10000,
        buffer_size: Optional[int] = None,
        read_timeout: Optional[int] = None,
        disconnection_timeout: Optional[int] = None,
        filesize_limit: int = 0,
        duration_limit: int = 0,
    ) -> None:
        super().__init__(
            live=live,
            out_dir=out_dir,
            path_template=path_template,
            stream_format='flv',
            recording_mode='standard',
            quality_number=quality_number,
            buffer_size=buffer_size,
            read_timeout=read_timeout,
            disconnection_timeout=disconnection_timeout,
            filesize_limit=filesize_limit,
            duration_limit=duration_limit,
        )
        self._init_for_debug(live.room_id)

        self._stream_fetcher = core_ops.StreamFetcher(
            live, self._session, read_timeout=read_timeout
        )

        self._prober = flv_ops.Prober()
        self._dl_statistics = core_ops.StreamStatistics()

        self._stream_parser = core_ops.StreamParser(self._stream_param_holder)
        self._analyser = flv_ops.Analyser()
        self._injector = flv_ops.Injector(self._metadata_provider)
        self._join_point_extractor = flv_ops.JoinPointExtractor()
        self._limiter = flv_ops.Limiter(filesize_limit, duration_limit)
        self._cutter = flv_ops.Cutter()
        self._dumper = flv_ops.Dumper(self._path_provider, buffer_size)
        self._metadata_dumper = MetadataDumper(
            self._dumper, self._analyser, self._join_point_extractor
        )

        self._recording_monitor = core_ops.RecordingMonitor(
            live, lambda: self._analyser.duration
        )

        self._prober.profiles.subscribe(self._on_profile_updated)
        self._dumper.file_opened.subscribe(self._on_video_file_opened)
        self._dumper.file_closed.subscribe(self._on_video_file_closed)
        self._recording_monitor.interrupted.subscribe(self._on_recording_interrupted)
        self._recording_monitor.recovered.subscribe(self._on_recording_recovered)

    @property
    def read_timeout(self) -> int:
        return self._stream_fetcher.read_timeout

    @read_timeout.setter
    def read_timeout(self, value: int) -> None:
        self._stream_fetcher.read_timeout = value

    @property
    def buffer_size(self) -> int:
        return self._dumper.buffer_size

    @buffer_size.setter
    def buffer_size(self, value: int) -> None:
        self._dumper.buffer_size = value

    @property
    def recording_path(self) -> Optional[str]:
        return self._dumper.path

    @property
    def filesize_limit(self) -> int:
        return self._limiter.filesize_limit

    @filesize_limit.setter
    def filesize_limit(self, value: int) -> None:
        self._limiter.filesize_limit = value

    @property
    def duration_limit(self) -> int:
        return self._limiter.duration_limit

    @duration_limit.setter
    def duration_limit(self, value: int) -> None:
        self._limiter.duration_limit = value

    @property
    def metadata(self) -> Optional[flv_ops.MetaData]:
        try:
            return self._analyser.make_metadata()
        except Exception:
            return None

    def can_cut_stream(self) -> bool:
        return self._cutter.can_cut_stream()

    def cut_stream(self) -> bool:
        return self._cutter.cut_stream()

    def _on_start(self) -> None:
        self._metadata_dumper.enable()

    def _on_stop(self) -> None:
        self._metadata_dumper.disable()

    def _run(self) -> None:
        self._subscription = (
            self._stream_param_holder.get_stream_params()  # type: ignore
            .pipe(
                self._stream_url_resolver,
                self._stream_fetcher,
                self._recording_monitor,
                self._dl_statistics,
                self._stream_parser,
                self._connection_error_handler,
                self._request_exception_handler,
                flv_ops.process(sort_tags=True),
                self._cutter,
                self._limiter,
                self._join_point_extractor,
                self._prober,
                self._injector,
                self._analyser,
                self._dumper,
                self._rec_statistics,
                self._progress_bar,
                self._exception_handler,
            )
            .subscribe(
                on_completed=self._on_completed,
                scheduler=NewThreadScheduler(self._thread_factory('StreamRecorder')),
            )
        )



from __future__ import annotations

import json
from subprocess import PIPE, Popen
from typing import Any, Dict, Optional

from reactivex import Observable, abc
from reactivex.scheduler import CurrentThreadScheduler

__all__ = ('ffprobe', 'StreamProfile')

StreamProfile = Dict[str, Any]


def ffprobe(data: bytes) -> Observable[StreamProfile]:
    def subscribe(
        observer: abc.ObserverBase[StreamProfile],
        scheduler: Optional[abc.SchedulerBase] = None,
    ) -> abc.DisposableBase:
        _scheduler = scheduler or CurrentThreadScheduler()

        def action(scheduler: abc.SchedulerBase, state: Optional[Any] = None) -> None:
            args = [
                'ffprobe',
                '-show_streams',
                '-show_format',
                '-print_format',
                'json',
                'pipe:0',
            ]

            with Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE) as process:
                try:
                    stdout, stderr = process.communicate(data, timeout=10)
                except Exception as e:
                    process.kill()
                    process.wait()
                    observer.on_error(e)
                else:
                    profile = json.loads(stdout)
                    observer.on_next(profile)
                    observer.on_completed()

        return _scheduler.schedule(action)

    return Observable(subscribe)



from typing import TypeVar

from pydantic.main import BaseModel


__all__ = 'update_settings', 'shadow_settings'


_T = TypeVar('_T', bound=BaseModel)


def update_settings(src: _T, dst: _T) -> None:
    overwrite_settings(src, dst, exclude_unset=True)


def shadow_settings(src: _T, dst: _T) -> None:
    overwrite_settings(src, dst, exclude_none=True)


def overwrite_settings(
    src: _T, dst: _T, exclude_unset: bool = False, exclude_none: bool = False
) -> None:
    # XXX Caution! the src and dst should be the same model type
    assert isinstance(src, BaseModel) and isinstance(dst, BaseModel)

    fields = src.__fields_set__ if exclude_unset else src.__fields__

    for name in fields:
        if not hasattr(dst, name):
            continue
        value = getattr(src, name)
        if exclude_none and value is None:
            continue
        if not isinstance(value, BaseModel):
            setattr(dst, name, value)
        else:
            overwrite_settings(
                value,
                getattr(dst, name),
                exclude_unset=exclude_unset,
                exclude_none=exclude_none,
            )



from __future__ import annotations
import logging
from typing import Set, Tuple, Type

import attr

from ..setting import WebHookSettings
from ..event import (
    LiveBeganEvent,
    LiveEndedEvent,
    RoomChangeEvent,
    RecordingStartedEvent,
    RecordingFinishedEvent,
    RecordingCancelledEvent,
    VideoFileCreatedEvent,
    VideoFileCompletedEvent,
    DanmakuFileCreatedEvent,
    DanmakuFileCompletedEvent,
    RawDanmakuFileCreatedEvent,
    RawDanmakuFileCompletedEvent,
    CoverImageDownloadedEvent,
    SpaceNoEnoughEvent,
    VideoPostprocessingCompletedEvent,
    PostprocessingCompletedEvent,
)
from ..event.typing import Event


__all__ = 'WebHook',


logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class WebHook:
    url: str
    event_types: Tuple[Type[Event], ...]
    receive_exception: bool

    @classmethod
    def from_settings(cls, settings: WebHookSettings) -> WebHook:
        types: Set[Type[Event]] = set()

        if settings.live_began:
            types.add(LiveBeganEvent)
        if settings.live_ended:
            types.add(LiveEndedEvent)
        if settings.room_change:
            types.add(RoomChangeEvent)
        if settings.recording_started:
            types.add(RecordingStartedEvent)
        if settings.recording_finished:
            types.add(RecordingFinishedEvent)
        if settings.recording_cancelled:
            types.add(RecordingCancelledEvent)
        if settings.video_file_created:
            types.add(VideoFileCreatedEvent)
        if settings.video_file_completed:
            types.add(VideoFileCompletedEvent)
        if settings.danmaku_file_created:
            types.add(DanmakuFileCreatedEvent)
        if settings.danmaku_file_completed:
            types.add(DanmakuFileCompletedEvent)
        if settings.raw_danmaku_file_created:
            types.add(RawDanmakuFileCreatedEvent)
        if settings.raw_danmaku_file_completed:
            types.add(RawDanmakuFileCompletedEvent)
        if settings.cover_image_downloaded:
            types.add(CoverImageDownloadedEvent)
        if settings.video_postprocessing_completed:
            types.add(VideoPostprocessingCompletedEvent)
        if settings.postprocessing_completed:
            types.add(PostprocessingCompletedEvent)
        if settings.space_no_enough:
            types.add(SpaceNoEnoughEvent)

        return cls(
            url=settings.url,
            event_types=tuple(types),
            receive_exception=settings.error_occurred,
        )



from .connection_error_handler import ConnectionErrorHandler
from .exception_handler import ExceptionHandler
from .progress_bar import ProgressBar
from .recording_monitor import RecordingMonitor
from .request_exception_handler import RequestExceptionHandler
from .sized_statistics import SizedStatistics
from .stream_fetcher import StreamFetcher
from .stream_parser import StreamParser
from .stream_statistics import StreamStatistics
from .stream_url_resolver import StreamURLResolver

__all__ = (
    'ConnectionErrorHandler',
    'ExceptionHandler',
    'ProgressBar',
    'RecordingMonitor',
    'RequestExceptionHandler',
    'SizedStatistics',
    'StreamFetcher',
    'StreamParser',
    'StreamStatistics',
    'StreamURLResolver',
)



import re
from typing import Optional


def snake_case(string: str) -> str:
    return re.sub(
        r'([a-z])([A-Z])', lambda m: m.group(1) + '_' + m.group(2).lower(), string
    )


def camel_case(string: str) -> str:
    words = string.split('_')
    return ''.join([words[0].casefold()] + [word.capitalize() for word in words[1:]])


def extract_uid_from_cookie(cookie: str) -> Optional[int]:
    match = re.search(r'DedeUserID=(\d+);', cookie)
    if match:
        return int(match.group(1))
    else:
        return None


def extract_buvid_from_cookie(cookie: str) -> Optional[str]:
    match = re.search(r'buvid3=([\w-]+);', cookie)
    if match:
        return match.group(1)
    else:
        return None



import struct


from .io_protocols import RandomIO


__all__ = 'StructReader', 'StructWriter'


class StructReader:
    def __init__(self, stream: RandomIO) -> None:
        self._stream = stream

    def read(self, size: int) -> bytes:
        data = self._stream.read(size)
        if len(data) != size:
            raise EOFError
        return data

    def read_ui8(self) -> int:
        return struct.unpack('B', self.read(1))[0]

    def read_ui16(self) -> int:
        return struct.unpack('>H', self.read(2))[0]

    def read_ui24(self) -> int:
        return struct.unpack('>I', b'\x00' + self.read(3))[0]

    def read_ui32(self) -> int:
        return struct.unpack('>I', self.read(4))[0]

    def read_si16(self) -> int:
        return struct.unpack('>h', self.read(2))[0]

    def read_f64(self) -> float:
        return struct.unpack('>d', self.read(8))[0]


class StructWriter:
    def __init__(self, stream: RandomIO) -> None:
        self._stream = stream

    def write(self, data: bytes) -> int:
        return self._stream.write(data)

    def write_ui8(self, number: int) -> int:
        return self.write(struct.pack('B', number))

    def write_ui16(self, number: int) -> int:
        return self.write(struct.pack('>H', number))

    def write_ui24(self, number: int) -> int:
        return self.write(struct.pack('>I', number)[1:])

    def write_ui32(self, number: int) -> int:
        return self.write(struct.pack('>I', number))

    def write_si16(self, number: int) -> int:
        return self.write(struct.pack('>h', number))

    def write_f64(self, number: float) -> int:
        return self.write(struct.pack('>d', number))



import logging
import os
from typing import Callable, List, Optional

from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

from ..common import (
    find_aac_header_tag,
    find_avc_header_tag,
    find_metadata_tag,
    is_audio_tag,
    is_avc_end_sequence,
    is_script_tag,
    is_video_nalu_keyframe,
    is_video_tag,
)
from ..models import AudioTag, FlvHeader, FlvTag, ScriptTag, VideoTag
from .typing import FLVStream, FLVStreamItem

__all__ = ('sort',)

logger = logging.getLogger(__name__)

TRACE_OP_SORT = bool(os.environ.get('BLREC_TRACE_OP_SORT'))


def sort() -> Callable[[FLVStream], FLVStream]:
    "Sort tags in GOP by timestamp to ensure subsequent operators work as expected."

    def _sort(source: FLVStream) -> FLVStream:
        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()

            gop_tags: List[FlvTag] = []

            def reset() -> None:
                nonlocal gop_tags
                gop_tags = []

            def push_gop_tags() -> None:
                if not gop_tags:
                    return

                if TRACE_OP_SORT:
                    logger.debug(
                        'Tags in GOP:\n'
                        f'Number of tags: {len(gop_tags)}\n'
                        f'Total size of tags: {sum(map(len, gop_tags))}\n'
                        f'The first tag is {gop_tags[0]}\n'
                        f'The last tag is {gop_tags[-1]}'
                    )

                if len(gop_tags) < 10:
                    avc_header_tag = find_avc_header_tag(gop_tags)
                    aac_header_tag = find_aac_header_tag(gop_tags)
                    if avc_header_tag is not None and aac_header_tag is not None:
                        if (metadata_tag := find_metadata_tag(gop_tags)) is not None:
                            observer.on_next(metadata_tag)
                        observer.on_next(avc_header_tag)
                        observer.on_next(aac_header_tag)
                        gop_tags.clear()
                        return

                script_tags: List[ScriptTag] = []
                video_tags: List[VideoTag] = []
                audio_tags: List[AudioTag] = []
                for tag in gop_tags:
                    if is_video_tag(tag):
                        video_tags.append(tag)
                    elif is_audio_tag(tag):
                        audio_tags.append(tag)
                    elif is_script_tag(tag):
                        script_tags.append(tag)

                sorted_tags: List[FlvTag] = []
                i = len(audio_tags) - 1
                for video_tag in reversed(video_tags):
                    sorted_tags.insert(0, video_tag)
                    while i >= 0 and audio_tags[i].timestamp >= video_tag.timestamp:
                        sorted_tags.insert(1, audio_tags[i])
                        i -= 1

                for tag in script_tags:
                    observer.on_next(tag)

                for tag in sorted_tags:
                    observer.on_next(tag)

                gop_tags.clear()

            def on_next(item: FLVStreamItem) -> None:
                if isinstance(item, FlvHeader) or is_avc_end_sequence(item):
                    push_gop_tags()
                    observer.on_next(item)
                    return

                if is_video_nalu_keyframe(item):
                    push_gop_tags()
                    gop_tags.append(item)
                else:
                    gop_tags.append(item)

            def on_completed() -> None:
                push_gop_tags()
                observer.on_completed()

            def on_error(exc: Exception) -> None:
                push_gop_tags()
                observer.on_error(exc)

            def dispose() -> None:
                nonlocal disposed
                disposed = True
                if gop_tags:
                    logger.debug(
                        'Remaining tags:\n'
                        f'Number of tags: {len(gop_tags)}\n'
                        f'Total size of tags: {sum(map(len, gop_tags))}\n'
                        f'The first tag is {gop_tags[0]}\n'
                        f'The last tag is {gop_tags[-1]}'
                    )
                reset()

            subscription.disposable = source.subscribe(
                on_next, on_error, on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    return _sort



import logging
from typing import Literal

import attr

from blrec.bili.typing import Danmaku

logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class DanmuMsg:
    mode: int
    size: int  # font size
    color: int
    date: int  # a timestamp in miliseconds
    dmid: int
    pool: int
    uid_hash: str
    uid: int
    uname: str  # sender name
    text: str

    @staticmethod
    def from_danmu(danmu: Danmaku) -> 'DanmuMsg':
        info = danmu['info']
        return DanmuMsg(
            mode=int(info[0][1]),
            size=int(info[0][2]),
            color=int(info[0][3]),
            date=int(info[0][4]),
            dmid=int(info[0][5]),
            pool=int(info[0][6]),
            uid_hash=info[0][7],
            uid=int(info[2][0]),
            uname=info[2][1],
            text=info[1],
        )


@attr.s(auto_attribs=True, slots=True, frozen=True)
class UserToastMsg:
    start_time: int  # timestamp in seconds
    uid: int
    username: str
    unit: str
    num: int
    price: int
    role_name: str
    guard_level: str
    toast_msg: str

    @staticmethod
    def from_danmu(danmu: Danmaku) -> 'UserToastMsg':
        data = danmu['data']
        return UserToastMsg(
            start_time=data['start_time'],
            uid=data['uid'],
            username=data['username'],
            unit=data['unit'],
            num=data['num'],
            price=data['price'],
            role_name=data['role_name'],
            guard_level=data['guard_level'],
            toast_msg=data['toast_msg'].replace('<%', '').replace('%>', ''),
        )


@attr.s(auto_attribs=True, frozen=True, slots=True)
class GiftSendMsg:
    gift_name: str
    count: int
    coin_type: Literal['sliver', 'gold']
    price: int
    uid: int
    uname: str
    timestamp: int  # timestamp in seconds

    @staticmethod
    def from_danmu(danmu: Danmaku) -> 'GiftSendMsg':
        data = danmu['data']
        return GiftSendMsg(
            gift_name=data['giftName'],
            count=int(data['num']),
            coin_type=data['coin_type'],
            price=int(data['price']),
            uid=int(data['uid']),
            uname=data['uname'],
            timestamp=int(data['timestamp']),
        )


@attr.s(auto_attribs=True, frozen=True, slots=True)
class GuardBuyMsg:
    gift_name: str
    count: int
    price: int
    uid: int
    uname: str
    guard_level: int  # 1 总督, 2 提督, 3 舰长
    timestamp: int  # timestamp in seconds

    @staticmethod
    def from_danmu(danmu: Danmaku) -> 'GuardBuyMsg':
        data = danmu['data']
        return GuardBuyMsg(
            gift_name=data['gift_name'],
            count=int(data['num']),
            price=int(data['price']),
            uid=int(data['uid']),
            uname=data['username'],
            guard_level=int(data['guard_level']),
            timestamp=int(data['start_time']),
        )


@attr.s(auto_attribs=True, frozen=True, slots=True)
class SuperChatMsg:
    gift_name: str
    count: int
    price: int
    rate: int
    time: int  # duration in seconds
    message: str
    uid: int
    uname: str
    timestamp: int  # timestamp in seconds

    @staticmethod
    def from_danmu(danmu: Danmaku) -> 'SuperChatMsg':
        data = danmu['data']
        return SuperChatMsg(
            gift_name=data['gift']['gift_name'],
            count=int(data['gift']['num']),
            price=int(data['price']),
            rate=int(data['rate']),
            time=int(data['time']),
            message=data['message'],
            uid=int(data['uid']),
            uname=data['user_info']['uname'],
            timestamp=int(data['ts']),
        )



import os
import errno

from fastapi import APIRouter, Body

from ..schemas import ResponseMessage
from ...application import Application


app: Application = None  # type: ignore  # bypass flake8 F821

router = APIRouter(
    prefix='/api/v1/validation',
    tags=['validation'],
)


@router.post(
    '/dir',
    response_model=ResponseMessage,
)
async def validate_dir(path: str = Body(..., embed=True)) -> ResponseMessage:
    """Check if the path is a directory and grants the read, write permissions
    """
    if not os.path.isdir(path):
        return ResponseMessage(code=errno.ENOTDIR, message='not a directory')
    elif not os.access(path, os.F_OK | os.R_OK | os.W_OK):
        return ResponseMessage(code=errno.EACCES, message='no permissions')
    else:
        return ResponseMessage(code=0, message='ok')



from __future__ import annotations

import io
import logging
from typing import Optional

from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

from blrec.flv.common import (
    is_audio_sequence_header,
    is_metadata_tag,
    is_video_sequence_header,
)
from blrec.flv.io import FlvReader
from blrec.flv.models import AudioTag, FlvHeader, ScriptTag, VideoTag
from blrec.flv.operators.typing import FLVStream, FLVStreamItem

__all__ = ('SegmentParser',)


logger = logging.getLogger(__name__)


class SegmentParser:
    def __init__(self) -> None:
        self._backup_timestamp = True

    def __call__(self, source: Observable[bytes]) -> FLVStream:
        return self._parse(source)

    def _parse(self, source: Observable[bytes]) -> FLVStream:
        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()

            last_flv_header: Optional[FlvHeader] = None
            last_metadata_tag: Optional[ScriptTag] = None
            last_audio_sequence_header: Optional[AudioTag] = None
            last_video_sequence_header: Optional[VideoTag] = None

            def reset() -> None:
                nonlocal last_flv_header, last_metadata_tag
                nonlocal last_audio_sequence_header, last_video_sequence_header
                last_flv_header = None
                last_metadata_tag = None
                last_audio_sequence_header = None
                last_video_sequence_header = None

            def on_next(data: bytes) -> None:
                nonlocal last_flv_header, last_metadata_tag
                nonlocal last_audio_sequence_header, last_video_sequence_header

                if b'' == data:
                    reset()
                    return

                try:
                    reader = FlvReader(
                        io.BytesIO(data), backup_timestamp=self._backup_timestamp
                    )

                    flv_header = reader.read_header()
                    if not last_flv_header:
                        observer.on_next(flv_header)
                        last_flv_header = flv_header
                    else:
                        assert last_flv_header == flv_header

                    while not disposed:
                        tag = reader.read_tag()
                        if is_metadata_tag(tag):
                            if last_metadata_tag is not None:
                                continue
                            last_metadata_tag = tag
                        elif is_video_sequence_header(tag):
                            if tag == last_video_sequence_header:
                                continue
                            last_video_sequence_header = tag
                        elif is_audio_sequence_header(tag):
                            if tag == last_audio_sequence_header:
                                continue
                            last_audio_sequence_header = tag
                        observer.on_next(tag)
                except EOFError:
                    pass
                except Exception as e:
                    observer.on_error(e)

            def dispose() -> None:
                nonlocal disposed
                disposed = True
                reset()

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)



from __future__ import annotations

import asyncio
import json
import logging
from decimal import Decimal
from typing import Iterable, List, Tuple, cast

import aiofiles
import m3u8

from blrec.flv.helpers import make_comment_for_joinpoints
from blrec.flv.operators import JoinPoint
from blrec.flv.utils import format_timestamp
from blrec.path.helpers import ffmpeg_metadata_path

from .helpers import get_extra_metadata, get_metadata, get_record_metadata

logger = logging.getLogger(__name__)


async def make_metadata_file(video_path: str) -> str:
    path = ffmpeg_metadata_path(video_path)
    async with aiofiles.open(path, 'wb') as file:
        if video_path.endswith('.flv'):
            content = await _make_metadata_content_for_flv(video_path)
        elif video_path.endswith('.m3u8'):
            content = await _make_metadata_content_for_m3u8(video_path)
        else:
            raise NotImplementedError(video_path)
        await file.write(content.encode(encoding='utf8'))
    return path


async def _make_metadata_content_for_flv(flv_path: str) -> str:
    metadata = await get_metadata(flv_path)
    try:
        extra_metadata = await get_extra_metadata(flv_path)
    except Exception as e:
        logger.warning(f'Failed to get extra metadata: {repr(e)}')
        extra_metadata = {}

    comment = cast(str, metadata.get('Comment', ''))
    chapters = ''

    if join_points := extra_metadata.get('joinpoints'):
        join_points = list(map(JoinPoint.from_metadata_value, join_points))
        comment += '\n\n' + make_comment_for_joinpoints(join_points)
        last_timestamp = int(
            cast(float, extra_metadata.get('duration') or metadata.get('duration'))
            * 1000
        )
        chapters = _make_chapters_for_flv(join_points, last_timestamp)

    comment = '\\\n'.join(comment.splitlines())

    # ref: https://ffmpeg.org/ffmpeg-formats.html#Metadata-1
    return f"""\
;FFMETADATA1
Title={metadata['Title']}
Artist={metadata['Artist']}
Date={metadata['Date']}
# Description may be truncated!
Description={json.dumps(metadata['description'], ensure_ascii=False)}
Comment={comment}

{chapters}
"""


def _make_chapters_for_flv(
    join_points: Iterable[JoinPoint], last_timestamp: int
) -> str:
    join_points = filter(lambda p: not p.seamless, join_points)
    timestamps = list(map(lambda p: p.timestamp, join_points))
    if not timestamps:
        return ''
    timestamps.insert(0, 0)
    timestamps.append(last_timestamp)

    result = ''
    for i in range(1, len(timestamps)):
        start = timestamps[i - 1]
        end = timestamps[i]
        if end < start:
            logger.warning(f'Chapter end time {end} before start {start}')
            end = start
        result += f"""\
[CHAPTER]
TIMEBASE=1/1000
START={start}
END={end}
title=segment \\#{i}
"""
    return result


async def _make_metadata_content_for_m3u8(playlist_path: str) -> str:
    metadata = await get_record_metadata(playlist_path)
    comment = cast(str, metadata.get('Comment', ''))
    chapters = ''

    timestamps, duration = await _get_discontinuities(playlist_path)
    if timestamps:
        comment += '\n\n' + _make_comment_for_discontinuities(timestamps)
        chapters = _make_chapters_for_m3u8(timestamps, duration)

    comment = '\\\n'.join(comment.splitlines())

    # ref: https://ffmpeg.org/ffmpeg-formats.html#Metadata-1
    return f"""\
;FFMETADATA1
Title={metadata['Title']}
Artist={metadata['Artist']}
Date={metadata['Date']}
# Description may be truncated!
Description={json.dumps(metadata['description'], ensure_ascii=False)}
Comment={comment}

{chapters}
"""


def _make_chapters_for_m3u8(timestamps: Iterable[int], duration: float) -> str:
    timestamps = list(timestamps)
    if not timestamps:
        return ''

    timestamps.insert(0, 0)
    timestamps.append(int(duration * 1000))

    result = ''
    for i in range(1, len(timestamps)):
        start = timestamps[i - 1]
        end = timestamps[i]
        if end < start:
            logger.warning(f'Chapter end time {end} before start {start}')
            end = start
        result += f"""\
[CHAPTER]
TIMEBASE=1/1000
START={start}
END={end}
title=segment \\#{i}
"""
    return result


def _make_comment_for_discontinuities(timestamps: Iterable[int]) -> str:
    return 'HLS片段不连续位置：\n' + '\n'.join(
        ('时间戳：{}'.format(format_timestamp(ts)) for ts in timestamps)
    )


async def _get_discontinuities(playlist_path: str) -> Tuple[List[int], float]:
    loop = asyncio.get_running_loop()
    playlist = await loop.run_in_executor(None, m3u8.load, playlist_path)
    duration = Decimal()
    timestamps: List[int] = []
    for seg in playlist.segments:
        if seg.discontinuity:
            timestamps.append(int(duration * 1000))
        duration += Decimal(str(seg.duration))
    return timestamps, float(duration)



from ctypes import cdll
from ctypes.util import find_library

lib_name = find_library('c')
if not lib_name:
    libc = None
else:
    libc = cdll.LoadLibrary(lib_name)


def malloc_trim(pad: int) -> bool:
    """Release free memory from the heap"""
    assert pad >= 0, 'pad must be >= 0'
    if libc is None:
        return False
    return libc.malloc_trim(pad) == 1



from __future__ import annotations

import json
import logging
from contextlib import suppress
from typing import Any, Callable, Dict, Tuple

from blrec.path.helpers import record_metadata_path
from blrec.utils.mixins import SwitchableMixin

from . import operators as hls_ops

__all__ = ('MetadataDumper',)

logger = logging.getLogger(__name__)


class MetadataDumper(SwitchableMixin):
    def __init__(
        self,
        playlist_dumper: hls_ops.PlaylistDumper,
        metadata_provider: Callable[[Dict[str, Any]], Dict[str, Any]],
    ) -> None:
        super().__init__()
        self._playlist_dumper = playlist_dumper
        self._metadata_provider = metadata_provider
        self._metadata: Dict[str, Any] = {}

    def _do_enable(self) -> None:
        self._file_opened_subscription = self._playlist_dumper.file_opened.subscribe(
            self._on_playlist_file_opened
        )
        logger.debug('Enabled metadata dumper')

    def _do_disable(self) -> None:
        with suppress(Exception):
            self._file_opened_subscription.dispose()
            del self._file_opened_subscription
        self._metadata.clear()
        logger.debug('Disabled metadata dumper')

    def _on_playlist_file_opened(self, args: Tuple[str, int]) -> None:
        playlist_path, _ = args
        metadata = self._metadata_provider({})
        self._dump_metadata(playlist_path, metadata)

    def _dump_metadata(self, playlist_path: str, metadata: Dict[str, Any]) -> None:
        path = record_metadata_path(playlist_path)
        logger.debug(f"Dumping metadata to file: '{path}'")

        with open(path, 'wt', encoding='utf8') as file:
            json.dump(metadata, file, ensure_ascii=False)



from __future__ import annotations

import logging
import os
from typing import Optional

import m3u8
from reactivex import Observable, abc
from reactivex import operators as ops
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

from blrec.core import operators as core_ops
from blrec.utils import operators as utils_ops

from ..exceptions import NoNewSegments

__all__ = ('PlaylistResolver',)


logger = logging.getLogger(__name__)


class PlaylistResolver:
    def __init__(self, stream_url_resolver: core_ops.StreamURLResolver) -> None:
        self._stream_url_resolver = stream_url_resolver

    def __call__(self, source: Observable[m3u8.M3U8]) -> Observable[m3u8.Segment]:
        return self._solve(source).pipe(
            ops.do_action(on_error=self._before_retry),
            utils_ops.retry(should_retry=self._should_retry),
        )

    def _name_of(self, uri: str) -> str:
        name, ext = os.path.splitext(uri)
        return name

    def _sequence_number_of(self, uri: str) -> int:
        return int(self._name_of(uri))

    def _solve(self, source: Observable[m3u8.M3U8]) -> Observable[m3u8.Segment]:
        def subscribe(
            observer: abc.ObserverBase[m3u8.Segment],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()

            attempts: int = 0
            last_sequence_number: Optional[int] = None

            def on_next(playlist: m3u8.M3U8) -> None:
                nonlocal attempts, last_sequence_number

                if playlist.is_endlist:
                    logger.debug('Playlist ended')

                new_segments = []
                for seg in playlist.segments:
                    num = self._sequence_number_of(seg.uri)
                    if last_sequence_number is not None:
                        if last_sequence_number >= num:
                            continue
                        if last_sequence_number + 1 != num:
                            logger.warning(
                                'Segments discontinuous: '
                                f'last sequence number: {last_sequence_number}, '
                                f'current sequence number: {num}'
                            )
                            seg.discontinuity = True
                    new_segments.append(seg)
                    last_sequence_number = num

                if not new_segments:
                    attempts += 1
                    if attempts > 3:
                        attempts = 0
                        observer.on_error(NoNewSegments())
                    return
                else:
                    attempts = 0

                for seg in new_segments:
                    observer.on_next(seg)

            def dispose() -> None:
                nonlocal disposed
                nonlocal last_sequence_number
                disposed = True
                last_sequence_number = None

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    def _should_retry(self, exc: Exception) -> bool:
        if isinstance(exc, NoNewSegments):
            return True
        else:
            return False

    def _before_retry(self, exc: Exception) -> None:
        if not isinstance(exc, NoNewSegments):
            return
        logger.warning('No new segments received, trying to update the stream url.')
        self._stream_url_resolver.reset()



from __future__ import annotations

import logging
from typing import Callable, Final, Optional, TypeVar

from reactivex import Observable, Subject, abc

from blrec.bili.live import Live
from blrec.utils.mixins import AsyncCooperationMixin

__all__ = ('RecordingMonitor',)


logger = logging.getLogger(__name__)

_T = TypeVar('_T')


class RecordingMonitor(AsyncCooperationMixin):
    def __init__(self, live: Live, duration_provider: Callable[..., float]) -> None:
        super().__init__()
        self._live = live
        self._duration_provider = duration_provider
        self._interrupted: Subject[float] = Subject()
        self._recovered: Subject[int] = Subject()

    @property
    def interrupted(self) -> Observable[float]:
        return self._interrupted

    @property
    def recovered(self) -> Observable[int]:
        return self._recovered

    def __call__(self, source: Observable[_T]) -> Observable[_T]:
        return self._monitor(source)

    def _monitor(self, source: Observable[_T]) -> Observable[_T]:
        CRITERIA: Final[int] = 1
        recording: bool = False
        failed_count: int = 0

        def subscribe(
            observer: abc.ObserverBase[_T],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            def on_next(item: _T) -> None:
                nonlocal recording, failed_count
                recording = True
                if failed_count >= CRITERIA:
                    ts = self._run_coroutine(self._live.get_timestamp())
                    self._recovered.on_next(ts)
                failed_count = 0
                observer.on_next(item)

            def on_error(exc: Exception) -> None:
                nonlocal failed_count
                if recording:
                    failed_count += 1
                    if failed_count == CRITERIA:
                        duration = self._duration_provider()
                        self._interrupted.on_next(duration)
                observer.on_error(exc)

            return source.subscribe(
                on_next, on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe)




from typing import Optional

import attr

from .io import DanmakuReader, DanmakuWriter


async def copy_damus(
    reader: DanmakuReader,
    writer: DanmakuWriter,
    *,
    timebase: Optional[int] = None,  # milliseconds
    delta: int = 0,  # milliseconds
) -> None:
    async for danmu in reader.read_danmus():
        if timebase is None:
            stime = max(0, danmu.stime * 1000 + delta) / 1000
        else:
            stime = max(0, danmu.date - timebase + delta) / 1000
        new_danmu = attr.evolve(danmu, stime=stime)
        await writer.write_danmu(new_danmu)



from __future__ import annotations
from abc import ABC
from contextlib import suppress
from typing import Any, List, Generic, TypeVar


from ..exception import ExceptionSubmitter


__all__ = 'EventListener', 'EventEmitter'


class EventListener(ABC):
    ...


_T = TypeVar('_T', bound=EventListener)


class EventEmitter(Generic[_T]):
    def __init__(self) -> None:
        super().__init__()
        self._listeners: List[_T] = []

    def add_listener(self, listener: _T) -> None:
        if listener not in self._listeners:
            self._listeners.append(listener)

    def remove_listener(self, listener: _T) -> None:
        with suppress(ValueError):
            self._listeners.remove(listener)

    # Python's type system is nowhere near Typescript's 😔
    # https://www.typescriptlang.org/docs/handbook/2/template-literal-types.html#string-unions-in-types
    # https://news.ycombinator.com/item?id=20824938
    # https://github.com/python/typing/issues/685
    async def _emit(self, name: str, *args: Any, **kwds: Any) -> None:
        for listener in self._listeners:
            with ExceptionSubmitter():
                coro = getattr(listener, 'on_' + name)
                await coro(*args, **kwds)



import asyncio
import json
import logging
import struct
from contextlib import suppress
from enum import Enum, IntEnum
from typing import Any, Dict, Final, List, Optional, Tuple, Union, cast

import aiohttp
import brotli
from aiohttp import ClientSession
from tenacity import retry, retry_if_exception_type, wait_exponential

from ..event.event_emitter import EventEmitter, EventListener
from ..exception import exception_callback
from ..logging.room_id import aio_task_with_room_id
from ..utils.mixins import AsyncStoppableMixin
from ..utils.string import extract_uid_from_cookie, extract_buvid_from_cookie
from .api import AppApi, WebApi
from .exceptions import DanmakuClientAuthError
from .typing import ApiPlatform, Danmaku

__all__ = 'DanmakuClient', 'DanmakuListener', 'Danmaku', 'DanmakuCommand'


logger = logging.getLogger(__name__)


class DanmakuListener(EventListener):
    async def on_client_connected(self) -> None:
        ...

    async def on_client_disconnected(self) -> None:
        ...

    async def on_client_reconnected(self) -> None:
        ...

    async def on_danmaku_received(self, danmu: Danmaku) -> None:
        ...

    async def on_error_occurred(self, error: Exception) -> None:
        ...


class DanmakuClient(EventEmitter[DanmakuListener], AsyncStoppableMixin):
    _HEARTBEAT_INTERVAL: Final[int] = 30

    def __init__(
        self,
        session: ClientSession,
        appapi: AppApi,
        webapi: WebApi,
        room_id: int,
        *,
        max_retries: int = 60,
        headers: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__()
        self.session = session
        self.appapi = appapi
        self.webapi = webapi
        self._uid = 0
        self._buvid = ''
        self._room_id = room_id
        self.headers = headers or {}

        self._api_platform: ApiPlatform = 'web'
        self._danmu_info: Dict[str, Any] = COMMON_DANMU_INFO
        self._host_index: int = 0
        self._retry_delay: int = 0
        self._MAX_RETRIES: Final[int] = max_retries

    @property
    def headers(self) -> Dict[str, str]:
        return self._headers

    @headers.setter
    def headers(self, value: Dict[str, str]) -> None:
        self._headers = {**value, 'Connection': 'Upgrade'}
        cookie = self._headers.get('Cookie', '')
        self._uid = extract_uid_from_cookie(cookie) or 0
        self._buvid = extract_buvid_from_cookie(cookie) or ''

    async def _do_start(self) -> None:
        await self._update_danmu_info()
        await self._connect()
        await self._create_message_loop()
        logger.debug('Started danmaku client')

    async def _do_stop(self) -> None:
        await self._terminate_message_loop()
        await self._disconnect()
        logger.debug('Stopped danmaku client')

    async def restart(self) -> None:
        logger.debug('Restarting danmaku client...')
        await self.stop()
        await self.start()
        logger.debug('Restarted danmaku client')

    async def reconnect(self) -> None:
        if self.stopped:
            return

        logger.debug('Reconnecting...')
        await self._disconnect()
        await self._connect()
        await self._emit('client_reconnected')

    @retry(
        wait=wait_exponential(multiplier=0.1, max=10),
        retry=retry_if_exception_type(
            (asyncio.TimeoutError, aiohttp.ClientError, ConnectionError)
        ),
    )
    async def _connect(self) -> None:
        logger.debug('Connecting to server...')
        try:
            await self._connect_websocket()
            await self._send_auth()
            reply = await self._recieve_auth_reply()
            await self._handle_auth_reply(reply)
        except Exception:
            self._host_index += 1
            if self._host_index >= len(self._danmu_info['host_list']):
                self._host_index = 0
                # self._rotate_api_platform()  # XXX: use web api only
                await self._update_danmu_info()
            raise
        else:
            logger.debug('Connected to server')
            await self._emit('client_connected')

    async def _connect_websocket(self) -> None:
        url = 'wss://{}:{}/sub'.format(
            self._danmu_info['host_list'][self._host_index]['host'],
            self._danmu_info['host_list'][self._host_index]['wss_port'],
        )
        logger.debug(f'Connecting WebSocket... {url}')
        try:
            self._ws = await self.session.ws_connect(
                url, timeout=5, headers=self.headers
            )
        except Exception as exc:
            logger.debug(f'Failed to connect WebSocket: {repr(exc)}')
            raise
        else:
            logger.debug('Connected WebSocket')

    async def _send_auth(self) -> None:
        auth_msg = json.dumps(
            {
                "uid": self._uid,
                'roomid': self._room_id,  # must not be the short id!
                'protover': WS.BODY_PROTOCOL_VERSION_BROTLI,
                "buvid": self._buvid,
                'platform': 'web',
                'type': 2,
                'key': self._danmu_info['token'],
            }
        )
        data = Frame.encode(WS.OP_USER_AUTHENTICATION, auth_msg)
        logger.debug('Sending user authentication...')
        try:
            await self._ws.send_bytes(data)
        except Exception as exc:
            logger.debug(f'Failed to sent user authentication: {repr(exc)}')
            raise
        else:
            logger.debug('Sent user authentication')

    async def _recieve_auth_reply(self) -> aiohttp.WSMessage:
        logger.debug('Receiving user authentication reply...')
        try:
            msg = await self._ws.receive(timeout=5)
            if msg.type != aiohttp.WSMsgType.BINARY:
                raise aiohttp.ClientError(msg)
        except Exception as exc:
            logger.debug(f'Failed to receive user authentication reply: {repr(exc)}')
            raise
        else:
            logger.debug('Recieved user authentication reply')
            return msg

    async def _handle_auth_reply(self, reply: aiohttp.WSMessage) -> None:
        op, msg = Frame.decode(reply.data)
        assert op == WS.OP_CONNECT_SUCCESS
        msg = cast(str, msg)
        code = cast(int, json.loads(msg)['code'])

        if code == WS.AUTH_OK:
            logger.debug('Auth OK')
            self._create_heartbeat_task()
        elif code == WS.AUTH_TOKEN_ERROR:
            raise DanmakuClientAuthError(f'Token expired: {code}')
        else:
            raise DanmakuClientAuthError(f'Unexpected code: {code}')

    def _rotate_api_platform(self) -> None:
        if self._api_platform == 'android':
            self._api_platform = 'web'
        else:
            self._api_platform = 'android'

    async def _update_danmu_info(self) -> None:
        logger.debug(f'Updating danmu info via {self._api_platform} api...')
        api: Union[WebApi, AppApi]
        if self._api_platform == 'web':
            api = self.webapi
        else:
            api = self.appapi
        try:
            self._danmu_info = await api.get_danmu_info(self._room_id)
        except Exception as exc:
            logger.warning(f'Failed to update danmu info: {repr(exc)}')
            self._danmu_info = COMMON_DANMU_INFO
        else:
            logger.debug('Danmu info updated')

    async def _disconnect(self) -> None:
        await self._cancel_heartbeat_task()
        await self._close_websocket()
        logger.debug('Disconnected from server')
        await self._emit('client_disconnected')

    async def _close_websocket(self) -> None:
        with suppress(BaseException):
            await self._ws.close()

    def _create_heartbeat_task(self) -> None:
        self._heartbeat_task = asyncio.create_task(self._send_heartbeat())
        self._heartbeat_task.add_done_callback(exception_callback)

    async def _cancel_heartbeat_task(self) -> None:
        self._heartbeat_task.cancel()
        with suppress(asyncio.CancelledError):
            await self._heartbeat_task

    @aio_task_with_room_id
    async def _send_heartbeat(self) -> None:
        data = Frame.encode(WS.OP_HEARTBEAT, '')
        while True:
            try:
                await self._ws.send_bytes(data)
            except Exception as exc:
                logger.warning(f'Failed to send heartbeat: {repr(exc)}')
                await self._emit('error_occurred', exc)
                task = asyncio.create_task(self.restart())
                task.add_done_callback(exception_callback)
                break
            await asyncio.sleep(self._HEARTBEAT_INTERVAL)

    async def _create_message_loop(self) -> None:
        self._message_loop_task = asyncio.create_task(self._message_loop())
        self._message_loop_task.add_done_callback(exception_callback)
        logger.debug('Created message loop')

    async def _terminate_message_loop(self) -> None:
        self._message_loop_task.cancel()
        with suppress(asyncio.CancelledError):
            await self._message_loop_task
        logger.debug('Terminated message loop')

    @aio_task_with_room_id
    async def _message_loop(self) -> None:
        while True:
            for msg in await self._receive():
                await self._dispatch_message(msg)

    async def _dispatch_message(self, msg: Dict[str, Any]) -> None:
        await self._emit('danmaku_received', msg)

    async def _receive(self) -> List[Dict[str, Any]]:
        self._reset_retry()

        while True:
            try:
                wsmsg = await self._ws.receive(timeout=self._HEARTBEAT_INTERVAL * 2)
            except Exception as e:
                await self._handle_receive_error(e)
            else:
                if wsmsg.type == aiohttp.WSMsgType.BINARY:
                    if result := await self._handle_data(wsmsg.data):
                        return result
                elif wsmsg.type == aiohttp.WSMsgType.ERROR:
                    await self._handle_receive_error(cast(Exception, wsmsg.data))
                elif wsmsg.type == aiohttp.WSMsgType.CLOSED:
                    msg = 'WebSocket Closed'
                    exc = aiohttp.WebSocketError(self._ws.close_code or 1006, msg)
                    await self._handle_receive_error(exc)
                else:
                    await self._handle_receive_error(ValueError(wsmsg))

    @staticmethod
    async def _handle_data(data: bytes) -> Optional[List[Dict[str, Any]]]:
        loop = asyncio.get_running_loop()

        try:
            op, msg = await loop.run_in_executor(None, Frame.decode, data)
            if op == WS.OP_MESSAGE:
                msg = cast(List[str], msg)
                return [json.loads(m) for m in msg]
            elif op == WS.OP_HEARTBEAT_REPLY:
                pass
        except Exception as e:
            logger.warning(f'Failed to handle data: {repr(e)}, data: {repr(data)}')

        return None

    async def _handle_receive_error(self, exc: Exception) -> None:
        logger.warning(f'Failed to receive message: {repr(exc)}')
        await self._emit('error_occurred', exc)
        if isinstance(exc, asyncio.TimeoutError):
            return
        await self._retry()

    def _reset_retry(self) -> None:
        self._retry_count = 0
        self._retry_delay = 0

    async def _retry(self) -> None:
        if self._retry_count < self._MAX_RETRIES:
            if self._retry_delay > 0:
                logger.debug(
                    'Retry after {} second{}'.format(
                        self._retry_delay, 's' if self._retry_delay > 1 else ''
                    )
                )
                await asyncio.sleep(self._retry_delay)
            await self.reconnect()
            self._retry_count += 1
            self._retry_delay += 1
        else:
            raise aiohttp.WebSocketError(1006, 'Over the maximum of retries')


class Frame:
    HEADER_FORMAT = '>IHHII'

    @staticmethod
    def encode(op: int, msg: str) -> bytes:
        body = msg.encode()
        header_length = WS.PACKAGE_HEADER_TOTAL_LENGTH
        packet_length = header_length + len(body)
        ver = WS.HEADER_DEFAULT_VERSION
        seq = WS.HEADER_DEFAULT_SEQUENCE

        header = struct.pack(
            Frame.HEADER_FORMAT,
            packet_length,
            header_length,
            ver,  # protocal version
            op,  # operation
            seq,  # sequence id
        )

        return header + body

    @staticmethod
    def decode(data: bytes) -> Tuple[int, Union[int, str, List[str]]]:
        plen, hlen, ver, op, _ = struct.unpack_from(Frame.HEADER_FORMAT, data, 0)
        body = data[hlen:]

        if op == WS.OP_MESSAGE:
            if ver == WS.BODY_PROTOCOL_VERSION_BROTLI:
                data = brotli.decompress(body)

            msg_list = []
            offset = 0
            while offset < len(data):
                plen, hlen, ver, op, _ = struct.unpack_from(
                    Frame.HEADER_FORMAT, data, offset
                )
                body = data[hlen + offset : plen + offset]
                msg = body.decode('utf8')
                msg_list.append(msg)
                offset += plen

            return op, msg_list
        elif op == WS.OP_HEARTBEAT_REPLY:
            online_count = cast(int, struct.unpack('>I', body)[0])
            return op, online_count
        elif op == WS.OP_CONNECT_SUCCESS:
            auth_result = body.decode()
            return op, auth_result
        else:
            raise ValueError(f'Unexpected Operation: {op}')


class WS(IntEnum):
    OP_HEARTBEAT = 2
    OP_HEARTBEAT_REPLY = 3
    OP_MESSAGE = 5
    OP_USER_AUTHENTICATION = 7
    OP_CONNECT_SUCCESS = 8
    PACKAGE_HEADER_TOTAL_LENGTH = 16
    PACKAGE_OFFSET = 0
    HEADER_OFFSET = 4
    VERSION_OFFSET = 6
    OPERATION_OFFSET = 8
    SEQUENCE_OFFSET = 12
    BODY_PROTOCOL_VERSION_NORMAL = 0
    BODY_PROTOCOL_VERSION_BROTLI = 3
    HEADER_DEFAULT_VERSION = 1
    HEADER_DEFAULT_OPERATION = 1
    HEADER_DEFAULT_SEQUENCE = 1
    AUTH_OK = 0
    AUTH_TOKEN_ERROR = -101


class DanmakuCommand(Enum):
    ACTIVITY_MATCH_GIFT = 'ACTIVITY_MATCH_GIFT'
    ANCHOR_LOT_AWARD = 'ANCHOR_LOT_AWARD'
    ANCHOR_LOT_CHECKSTATUS = 'ANCHOR_LOT_CHECKSTATUS'
    ANCHOR_LOT_END = 'ANCHOR_LOT_END'
    ANCHOR_LOT_START = 'ANCHOR_LOT_START'
    ANIMATION = 'ANIMATION'
    BOX_ACTIVITY_START = 'BOX_ACTIVITY_START'
    CHANGE_ROOM_INFO = 'CHANGE_ROOM_INFO'
    CHASE_FRAME_SWITCH = 'CHASE_FRAME_SWITCH'
    COMBO_SEND = 'COMBO_SEND'
    CUT_OFF = 'CUT_OFF'
    DANMU_GIFT_LOTTERY_AWARD = 'DANMU_GIFT_LOTTERY_AWARD'
    DANMU_GIFT_LOTTERY_END = 'DANMU_GIFT_LOTTERY_END'
    DANMU_GIFT_LOTTERY_START = 'DANMU_GIFT_LOTTERY_START'
    DANMU_MSG = 'DANMU_MSG'
    ENTRY_EFFECT = 'ENTRY_EFFECT'
    GUARD_ACHIEVEMENT_ROOM = 'GUARD_ACHIEVEMENT_ROOM'
    GUARD_LOTTERY_START = 'GUARD_LOTTERY_START'
    HOUR_RANK_AWARDS = 'HOUR_RANK_AWARDS'
    LITTLE_TIPS = 'LITTLE_TIPS'
    LIVE = 'LIVE'
    LOL_ACTIVITY = 'LOL_ACTIVITY'
    LUCK_GIFT_AWARD_USER = 'LUCK_GIFT_AWARD_USER'
    MATCH_TEAM_GIFT_RANK = 'MATCH_TEAM_GIFT_RANK'
    MESSAGEBOX_USER_GAIN_MEDAL = 'MESSAGEBOX_USER_GAIN_MEDAL'
    NOTICE_MSG = 'NOTICE_MSG'
    PK_AGAIN = 'PK_AGAIN'
    PK_BATTLE_CRIT = 'PK_BATTLE_CRIT'
    PK_BATTLE_END = 'PK_BATTLE_END'
    PK_BATTLE_GIFT = 'PK_BATTLE_GIFT'
    PK_BATTLE_PRE = 'PK_BATTLE_PRE'
    PK_BATTLE_PRO_TYPE = 'PK_BATTLE_PRO_TYPE'
    PK_BATTLE_PROCESS = 'PK_BATTLE_PROCESS'
    PK_BATTLE_RANK_CHANGE = 'PK_BATTLE_RANK_CHANGE'
    PK_BATTLE_SETTLE_USER = 'PK_BATTLE_SETTLE_USER'
    PK_BATTLE_SPECIAL_GIFT = 'PK_BATTLE_SPECIAL_GIFT'
    PK_BATTLE_START = 'PK_BATTLE_START'
    PK_BATTLE_VOTES_ADD = 'PK_BATTLE_VOTES_ADD'
    PK_END = 'PK_END'
    PK_LOTTERY_START = 'PK_LOTTERY_START'
    PK_MATCH = 'PK_MATCH'
    PK_MIC_END = 'PK_MIC_END'
    PK_PRE = 'PK_PRE'
    PK_PROCESS = 'PK_PROCESS'
    PK_SETTLE = 'PK_SETTLE'
    PK_START = 'PK_START'
    PREPARING = 'PREPARING'
    RAFFLE_END = 'RAFFLE_END'
    RAFFLE_START = 'RAFFLE_START'
    ROOM_BLOCK_INTO = 'ROOM_BLOCK_INTO'
    ROOM_BLOCK_MSG = 'ROOM_BLOCK_MSG'
    ROOM_BOX_USER = 'ROOM_BOX_USER'
    ROOM_CHANGE = 'ROOM_CHANGE'
    ROOM_KICKOUT = 'ROOM_KICKOUT'
    ROOM_LIMIT = 'ROOM_LIMIT'
    ROOM_LOCK = 'ROOM_LOCK'
    ROOM_RANK = 'ROOM_RANK'
    ROOM_REAL_TIME_MESSAGE_UPDATE = 'ROOM_REAL_TIME_MESSAGE_UPDATE'
    ROOM_REFRESH = 'ROOM_REFRESH'
    ROOM_SILENT_OFF = 'ROOM_SILENT_OFF'
    ROOM_SILENT_ON = 'ROOM_SILENT_ON'
    ROOM_SKIN_MSG = 'ROOM_SKIN_MSG'
    SCORE_CARD = 'SCORE_CARD'
    SEND_GIFT = 'SEND_GIFT'
    SEND_TOP = 'SEND_TOP'
    SPECIAL_GIFT = 'SPECIAL_GIFT'
    SUPER_CHAT_ENTRANCE = 'SUPER_CHAT_ENTRANCE'
    SUPER_CHAT_MESSAGE = 'SUPER_CHAT_MESSAGE'
    SUPER_CHAT_MESSAGE_DELETE = 'SUPER_CHAT_MESSAGE_DELETE'
    TV_END = 'TV_END'
    TV_START = 'TV_START'
    USER_TOAST_MSG = 'USER_TOAST_MSG'
    VOICE_JOIN_STATUS = 'VOICE_JOIN_STATUS'
    WARNING = 'WARNING'
    WATCH_LPL_EXPIRED = 'WATCH_LPL_EXPIRED'
    WEEK_STAR_CLOCK = 'WEEK_STAR_CLOCK'
    WELCOME = 'WELCOME'
    WELCOME_GUARD = 'WELCOME_GUARD'
    WIN_ACTIVITY = 'WIN_ACTIVITY'
    WIN_ACTIVITY_USER = 'WIN_ACTIVITY_USER'
    WISH_BOTTLE = 'WISH_BOTTLE'
    GUARD_BUY = 'GUARD_BUY'
    # ...


COMMON_DANMU_INFO: Final[Dict[str, Any]] = {
    "token": "",
    "host_list": [
        {
            "host": "broadcastlv.chat.bilibili.com",
            "port": 2243,
            "wss_port": 443,
            "ws_port": 2244,
        }
    ],
}



from __future__ import annotations

import io
import logging
from typing import List, Optional, cast

from reactivex import Observable, Subject, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

from ...utils.ffprobe import StreamProfile, ffprobe
from ..common import find_aac_header_tag, find_avc_header_tag
from ..io import FlvWriter
from ..models import FlvHeader, FlvTag
from .typing import FLVStream, FLVStreamItem

__all__ = ('Prober', 'StreamProfile')


logger = logging.getLogger(__name__)


class Prober:
    def __init__(self) -> None:
        self._profiles: Subject[StreamProfile] = Subject()

    def _reset(self) -> None:
        self._gathering: bool = False
        self._gathered_items: List[FLVStreamItem] = []

    @property
    def profiles(self) -> Observable[StreamProfile]:
        return self._profiles

    def __call__(self, source: FLVStream) -> FLVStream:
        return self._probe(source)

    def _probe(self, source: FLVStream) -> FLVStream:
        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()

            self._reset()

            def on_next(item: FLVStreamItem) -> None:
                if isinstance(item, FlvHeader):
                    self._gathered_items.clear()
                    self._gathering = True

                if self._gathering:
                    self._gathered_items.append(item)
                    if len(self._gathered_items) >= 10:
                        try:
                            self._do_probe()
                        except Exception as e:
                            logger.warning(f'Failed to probe stream: {repr(e)}')
                        finally:
                            self._gathered_items.clear()
                            self._gathering = False

                observer.on_next(item)

            def dispose() -> None:
                nonlocal disposed
                disposed = True
                self._reset()

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    def _do_probe(self) -> None:
        bytes_io = io.BytesIO()
        writer = FlvWriter(bytes_io)
        writer.write_header(cast(FlvHeader, self._gathered_items[0]))
        gathered_tags = cast(List[FlvTag], self._gathered_items[1:])

        avc_sequence_header = find_avc_header_tag(gathered_tags)
        aac_sequence_header = find_aac_header_tag(gathered_tags)
        if avc_sequence_header is not None and aac_sequence_header is not None:
            index_of_avc_sequence_header = gathered_tags.index(avc_sequence_header)
            index_of_aac_sequence_header = gathered_tags.index(aac_sequence_header)
            if index_of_avc_sequence_header > index_of_aac_sequence_header:
                gathered_tags[index_of_aac_sequence_header] = avc_sequence_header
                gathered_tags[index_of_avc_sequence_header] = aac_sequence_header

        for tag in gathered_tags:
            writer.write_tag(tag)

        def on_next(profile: StreamProfile) -> None:
            self._profiles.on_next(profile)

        def on_error(e: Exception) -> None:
            logger.warning(f'Failed to probe stream by ffprobe: {repr(e)}')

        ffprobe(bytes_io.getvalue()).subscribe(on_next, on_error)



from __future__ import annotations

import io
import logging
import os
from copy import deepcopy
from decimal import Decimal
from typing import Callable, Optional, Tuple

import m3u8
from reactivex import Observable, Subject, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

__all__ = ('PlaylistDumper',)

logger = logging.getLogger(__name__)


class PlaylistDumper:
    def __init__(self, path_provider: Callable[..., Tuple[str, int]]) -> None:
        self._path_provider = path_provider
        self._file_opened: Subject[Tuple[str, int]] = Subject()
        self._file_closed: Subject[str] = Subject()
        self._reset()

    def _reset(self) -> None:
        self._path: str = ''
        self._file: Optional[io.TextIOWrapper] = None
        self._duration: Decimal = Decimal()

    @property
    def path(self) -> str:
        return self._path

    @property
    def duration(self) -> float:
        return float(self._duration)

    @property
    def file_opened(self) -> Observable[Tuple[str, int]]:
        return self._file_opened

    @property
    def file_closed(self) -> Observable[str]:
        return self._file_closed

    def __call__(self, source: Observable[m3u8.M3U8]) -> Observable[m3u8.Segment]:
        return self._dump(source)

    def _open_file(self) -> None:
        path, timestamp = self._path_provider()
        root, ext = os.path.splitext(path)
        os.makedirs(root, exist_ok=True)
        self._path = os.path.join(root, 'index.m3u8')
        self._file = open(self._path, 'wt', encoding='utf8')  # type: ignore
        logger.debug(f'Opened file: {self._path}')
        self._file_opened.on_next((self._path, timestamp))

    def _close_file(self) -> None:
        if self._file is not None and not self._file.closed:
            self._file.write('#EXT-X-ENDLIST')
            self._file.close()
            logger.debug(f'Closed file: {self._path}')
            self._file_closed.on_next(self._path)

    def _name_of(self, uri: str) -> str:
        name, ext = os.path.splitext(uri)
        return name

    def _sequence_number_of(self, uri: str) -> int:
        return int(self._name_of(uri))

    def _replace_uri(self, segment: m3u8.Segment) -> m3u8.Segment:
        copied_seg = deepcopy(segment)
        if init_section := getattr(copied_seg, 'init_section', None):
            init_section.uri = f'segments/{init_section.uri}'
        uri = segment.uri
        name = self._name_of(uri)
        copied_seg.uri = 'segments/%s/%s' % (name[:-3], uri)
        return copied_seg

    def _replace_all_uri(self, playlist: m3u8.M3U8) -> m3u8.M3U8:
        copied_playlist = deepcopy(playlist)
        copied_playlist.segments = m3u8.SegmentList(
            self._replace_uri(s) for s in copied_playlist.segments
        )
        return copied_playlist

    def _update_duration(self, segment: m3u8.Segment) -> None:
        self._duration += Decimal(str(segment.duration))

    def _dump(self, source: Observable[m3u8.M3U8]) -> Observable[m3u8.Segment]:
        def subscribe(
            observer: abc.ObserverBase[m3u8.Segment],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()

            last_segment: Optional[m3u8.Segment] = None
            last_sequence_number: Optional[int] = None
            first_playlist_dumped: bool = False

            self._close_file()
            self._reset()

            def on_next(playlist: m3u8.M3U8) -> None:
                nonlocal last_sequence_number, last_segment, first_playlist_dumped

                if playlist.is_endlist:
                    logger.debug('Playlist ended')

                try:
                    if not first_playlist_dumped:
                        self._close_file()
                        self._reset()
                        self._open_file()
                        assert self._file is not None
                        playlist.is_endlist = False
                        self._file.write(self._replace_all_uri(playlist).dumps())
                        self._file.flush()
                        for seg in playlist.segments:
                            observer.on_next(seg)
                            self._update_duration(seg)
                            last_segment = seg
                            last_sequence_number = self._sequence_number_of(seg.uri)
                        first_playlist_dumped = True
                        logger.debug('The first playlist has been dumped')
                        return

                    assert self._file is not None
                    for seg in playlist.segments:
                        num = self._sequence_number_of(seg.uri)
                        discontinuity = False
                        if last_sequence_number is not None:
                            if last_sequence_number >= num:
                                continue
                            if last_sequence_number + 1 != num:
                                logger.warning(
                                    'Segments discontinuous: '
                                    f'last sequence number: {last_sequence_number}, '
                                    f'current sequence number: {num}'
                                )
                                discontinuity = True
                        new_seg = self._replace_uri(seg)
                        new_seg.discontinuity = discontinuity
                        new_last_seg = self._replace_uri(last_segment)
                        self._file.write(new_seg.dumps(new_last_seg) + '\n')
                        observer.on_next(seg)
                        self._update_duration(seg)
                        last_segment = seg
                        last_sequence_number = num
                except Exception as e:
                    self._close_file()
                    self._reset()
                    observer.on_error(e)

            def on_completed() -> None:
                self._close_file()
                self._reset()
                observer.on_completed()

            def on_error(e: Exception) -> None:
                self._close_file()
                self._reset()
                observer.on_error(e)

            def dispose() -> None:
                nonlocal disposed
                nonlocal last_segment, last_sequence_number
                disposed = True
                last_segment = None
                last_sequence_number = None
                self._close_file()
                self._reset()

            subscription.disposable = source.subscribe(
                on_next, on_error, on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)



from __future__ import annotations

import asyncio
import html
import logging
import unicodedata
from datetime import datetime, timedelta, timezone
from typing import Any, AsyncIterator, Final, List

import aiofiles
import attr
from lxml import etree

from .models import (
    Danmu,
    GiftSendRecord,
    GuardBuyRecord,
    Metadata,
    SuperChatRecord,
    UserToast,
)
from .typing import Element

__all__ = 'DanmakuReader', 'DanmakuWriter'


logger = logging.getLogger(__name__)


class DanmakuReader:  # TODO rewrite
    def __init__(self, path: str) -> None:
        self._path = path

    async def __aenter__(self) -> DanmakuReader:
        await self.init()
        return self

    async def __aexit__(self, exc_type, exc, tb):  # type: ignore
        pass

    async def init(self) -> None:
        loop = asyncio.get_running_loop()
        self._tree = await loop.run_in_executor(None, etree.parse, self._path)

    async def read_metadata(self) -> Metadata:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self._read_metadata)

    async def read_danmus(self) -> AsyncIterator[Danmu]:
        for elem in await self._get_danmu_elems():
            yield self._make_danmu(elem)

    def _read_metadata(self) -> Metadata:
        return Metadata(
            user_name=self._tree.xpath('/i/metadata/user_name')[0].text,
            room_id=int(self._tree.xpath('/i/metadata/room_id')[0].text),
            room_title=self._tree.xpath('/i/metadata/room_title')[0].text,
            area=self._tree.xpath('/i/metadata/area')[0].text,
            parent_area=self._tree.xpath('/i/metadata/parent_area')[0].text,
            live_start_time=int(
                datetime.fromisoformat(
                    self._tree.xpath('/i/metadata/live_start_time')[0].text
                ).timestamp()
            ),
            record_start_time=int(
                datetime.fromisoformat(
                    self._tree.xpath('/i/metadata/record_start_time')[0].text
                ).timestamp()
            ),
            recorder=self._tree.xpath('/i/metadata/recorder')[0].text,
        )

    async def _get_danmu_elems(self) -> List[Element]:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self._tree.xpath, '/i/d')

    def _make_danmu(self, elem: Element) -> Danmu:
        params = elem.get('p').split(',')
        return Danmu(
            stime=float(params[0]),
            mode=int(params[1]),
            size=int(params[2]),
            color=int(params[3]),
            date=int(params[4]),
            pool=int(params[5]),
            uid_hash=params[6],
            uid=elem.get('uid'),
            uname=elem.get('user'),
            dmid=int(params[7]),
            text=elem.text,
        )


class DanmakuWriter:
    _XML_HEAD: Final[
        str
    ] = """\
<?xml version="1.0" encoding="UTF-8"?>
<i>
    <chatserver>chat.bilibili.com</chatserver>
    <chatid>0</chatid>
    <mission>0</mission>
    <maxlimit>0</maxlimit>
    <state>0</state>
    <real_name>0</real_name>
    <source>e-r</source>
"""

    def __init__(self, path: str):
        self._path = path

    async def __aenter__(self) -> DanmakuWriter:
        await self.init()
        return self

    async def __aexit__(self, exc_type, exc, tb):  # type: ignore
        await self.complete()

    async def init(self) -> None:
        self._file = await aiofiles.open(self._path, 'wt', encoding='utf8')
        await self._file.write(self._XML_HEAD)

    async def write_metadata(self, metadata: Metadata) -> None:
        await self._file.write(self._serialize_metadata(metadata))

    async def write_danmu(self, danmu: Danmu) -> None:
        await self._file.write(self._serialize_danmu(danmu))

    async def write_user_toast(self, toast: UserToast) -> None:
        await self._file.write(self._serialize_user_toast(toast))

    async def write_gift_send_record(self, record: GiftSendRecord) -> None:
        await self._file.write(self._serialize_gift_send_record(record))

    async def write_guard_buy_record(self, record: GuardBuyRecord) -> None:
        await self._file.write(self._serialize_guard_buy_record(record))

    async def write_super_chat_record(self, record: SuperChatRecord) -> None:
        await self._file.write(self._serialize_super_chat_record(record))

    async def complete(self) -> None:
        await self._file.write('</i>')
        await self._file.close()

    def _serialize_metadata(self, metadata: Metadata) -> str:
        tz = timezone(timedelta(hours=8))
        ts = metadata.live_start_time
        live_start_time = datetime.fromtimestamp(ts, tz).isoformat()
        ts = metadata.record_start_time
        record_start_time = datetime.fromtimestamp(ts, tz).isoformat()
        return f"""\
    <metadata>
        <user_name>{html.escape(metadata.user_name)}</user_name>
        <room_id>{metadata.room_id}</room_id>
        <room_title>{html.escape(metadata.room_title)}</room_title>
        <area>{html.escape(metadata.area)}</area>
        <parent_area>{html.escape(metadata.parent_area)}</parent_area>
        <live_start_time>{live_start_time}</live_start_time>
        <record_start_time>{record_start_time}</record_start_time>
        <recorder>{metadata.recorder}</recorder>
    </metadata>
"""

    def _serialize_danmu(self, dm: Danmu) -> str:
        attrib = {
            'p': (
                f'{dm.stime:.3f},{dm.mode},{dm.size},{dm.color},'
                f'{dm.date},{dm.pool},{dm.uid_hash},{dm.dmid}'
            ),
            'uid': str(dm.uid),
            'user': dm.uname,
        }

        try:
            elem = etree.Element('d', attrib=attrib)
        except ValueError:
            # ValueError: All strings must be XML compatible: Unicode or ASCII,
            # no NULL bytes or control characters
            attrib['user'] = remove_control_characters(dm.uname)
            elem = etree.Element('d', attrib=attrib)

        try:
            elem.text = dm.text
        except ValueError:
            # ValueError: All strings must be XML compatible: Unicode or ASCII,
            # no NULL bytes or control characters
            elem.text = remove_control_characters(dm.text)

        return '    ' + etree.tostring(elem, encoding='utf8').decode() + '\n'

    def _serialize_user_toast(self, toast: UserToast) -> str:
        attrib = attr.asdict(
            toast,
            filter=lambda a, v: a.name != 'msg',
            value_serializer=record_value_serializer,
        )
        elem = etree.Element('toast', attrib=attrib)
        elem.text = remove_control_characters(toast.msg)
        return '    ' + etree.tostring(elem, encoding='utf8').decode() + '\n'

    def _serialize_gift_send_record(self, record: GiftSendRecord) -> str:
        attrib = attr.asdict(record, value_serializer=record_value_serializer)
        elem = etree.Element('gift', attrib=attrib)
        return '    ' + etree.tostring(elem, encoding='utf8').decode() + '\n'

    def _serialize_guard_buy_record(self, record: GuardBuyRecord) -> str:
        attrib = attr.asdict(record, value_serializer=record_value_serializer)
        elem = etree.Element('guard', attrib=attrib)
        return '    ' + etree.tostring(elem, encoding='utf8').decode() + '\n'

    def _serialize_super_chat_record(self, record: SuperChatRecord) -> str:
        attrib = attr.asdict(
            record,
            filter=lambda a, v: a.name != 'message',
            value_serializer=record_value_serializer,
        )
        elem = etree.Element('sc', attrib=attrib)
        elem.text = remove_control_characters(record.message)
        return '    ' + etree.tostring(elem, encoding='utf8').decode() + '\n'


def record_value_serializer(
    instance: Any, attribute: attr.Attribute[Any], value: Any
) -> Any:
    if attribute.name == 'ts':
        return f'{value:.3f}'
    if attribute.name == 'cointype':
        return '金瓜子' if value == 'gold' else '银瓜子'
    if not isinstance(value, str):
        value = str(value)
    return remove_control_characters(value)


def remove_control_characters(s: str) -> str:
    return ''.join(c for c in s if unicodedata.category(c) != 'Cc')



import os
import shutil
import logging
import asyncio
from typing import Literal


logger = logging.getLogger(__name__)


def is_space_enough(path: str, size: int) -> bool:
    return shutil.disk_usage(path).free > size


async def delete_file(
    path: str, log_level: Literal['INFO', 'DEBUG'] = 'INFO'
) -> None:
    loop = asyncio.get_running_loop()

    try:
        await loop.run_in_executor(None, os.remove, path)
    except Exception as e:
        logger.error(f'Failed to delete {path!r}, due to: {repr(e)}')
    else:
        logger.log(logging.getLevelName(log_level), f'Deleted {path!r}')



from queue import Queue
from threading import Thread
from typing import Any, Callable, Optional, TypeVar

from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable


_T = TypeVar('_T')


def observe_on_new_thread(
    queue_size: Optional[int] = None, thread_name: Optional[str] = None
) -> Callable[[Observable[_T]], Observable[_T]]:
    def observe_on(source: Observable[_T]) -> Observable[_T]:
        def subscribe(
            observer: abc.ObserverBase[_T],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()
            queue: Queue[Callable[..., Any]] = Queue(maxsize=queue_size or 0)

            def run() -> None:
                while not disposed:
                    queue.get()()

            thread = Thread(target=run, name=thread_name, daemon=True)
            thread.start()

            def on_next(value: _T) -> None:
                queue.put(lambda: observer.on_next(value))

            def on_error(exc: Exception) -> None:
                queue.put(lambda: observer.on_error(exc))

            def on_completed() -> None:
                queue.put(lambda: observer.on_completed)

            def dispose() -> None:
                nonlocal disposed
                disposed = True
                queue.put(lambda: None)
                thread.join()

            subscription.disposable = source.subscribe(
                on_next, on_error, on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    return observe_on



import logging
import os
from typing import Optional, Tuple

from brotli_asgi import BrotliMiddleware
from fastapi import Depends, FastAPI, Request, status
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from pkg_resources import resource_filename
from pydantic import ValidationError
from starlette.responses import Response

from blrec.exception import ExistsError, ForbiddenError, NotFoundError
from blrec.path.helpers import create_file, file_exists
from blrec.setting import EnvSettings, Settings
from blrec.web.middlewares.base_herf import BaseHrefMiddleware
from blrec.web.middlewares.route_redirect import RouteRedirectMiddleware

from ..application import Application
from . import security
from .routers import application, settings, tasks, update, validation, websockets
from .schemas import ResponseMessage

logger = logging.getLogger(__name__)


_env_settings = EnvSettings()
_path = os.path.abspath(os.path.expanduser(_env_settings.settings_file))
if not file_exists(_path):
    create_file(_path)
_env_settings.settings_file = _path

_settings = Settings.load(_env_settings.settings_file)
_settings.update_from_env_settings(_env_settings)

app = Application(_settings)

if _env_settings.api_key is None:
    _dependencies = None
else:
    security.api_key = _env_settings.api_key
    _dependencies = [Depends(security.authenticate)]

api = FastAPI(
    title='Bilibili live streaming recorder web API',
    description='Web API to communicate with the backend application',
    version='v1',
    dependencies=_dependencies,
)

api.add_middleware(BaseHrefMiddleware)
api.add_middleware(BrotliMiddleware)
api.add_middleware(
    CORSMiddleware,
    allow_origins=['http://localhost:4200'],  # angular development
    allow_credentials=True,
    allow_methods=['*'],
    allow_headers=['*'],
)
api.add_middleware(RouteRedirectMiddleware)


@api.exception_handler(NotFoundError)
async def not_found_error_handler(request: Request, exc: NotFoundError) -> JSONResponse:
    return JSONResponse(
        status_code=status.HTTP_404_NOT_FOUND,
        content=dict(ResponseMessage(code=status.HTTP_404_NOT_FOUND, message=str(exc))),
    )


@api.exception_handler(ForbiddenError)
async def forbidden_error_handler(
    request: Request, exc: ForbiddenError
) -> JSONResponse:
    return JSONResponse(
        status_code=status.HTTP_403_FORBIDDEN,
        content=dict(ResponseMessage(code=status.HTTP_403_FORBIDDEN, message=str(exc))),
    )


@api.exception_handler(ExistsError)
async def exists_error_handler(request: Request, exc: ExistsError) -> JSONResponse:
    return JSONResponse(
        status_code=status.HTTP_409_CONFLICT,
        content=dict(ResponseMessage(code=status.HTTP_409_CONFLICT, message=str(exc))),
    )


@api.exception_handler(ValidationError)
async def validation_error_handler(
    request: Request, exc: ValidationError
) -> JSONResponse:
    return JSONResponse(
        status_code=status.HTTP_406_NOT_ACCEPTABLE,
        content=dict(
            ResponseMessage(code=status.HTTP_406_NOT_ACCEPTABLE, message=str(exc))
        ),
    )


@api.on_event('startup')
async def on_startup() -> None:
    await app.launch()


@api.on_event('shutdown')
async def on_shuntdown() -> None:
    _settings.dump()
    await app.exit()


tasks.app = app
settings.app = app
application.app = app
validation.app = app
websockets.app = app
update.app = app
api.include_router(tasks.router)
api.include_router(settings.router)
api.include_router(application.router)
api.include_router(validation.router)
api.include_router(websockets.router)
api.include_router(update.router)


class WebAppFiles(StaticFiles):
    def lookup_path(self, path: str) -> Tuple[str, Optional[os.stat_result]]:
        if path == '404.html':
            path = 'index.html'
        return super().lookup_path(path)

    def file_response(self, full_path: str, *args, **kwargs) -> Response:  # type: ignore # noqa
        # ignore MIME types from Windows registry
        # workaround for https://github.com/acgnhiki/blrec/issues/12
        response = super().file_response(full_path, *args, **kwargs)
        if full_path.endswith('.js'):
            js_media_type = 'application/javascript'
            if response.media_type != js_media_type:
                response.media_type = js_media_type
                headers = response.headers
                headers['content-type'] = js_media_type
                response.raw_headers = headers.raw
                del response._headers
        return response


directory = resource_filename(__name__, '../data/webapp')
api.mount('/', WebAppFiles(directory=directory, html=True), name='webapp')



from abc import ABC
from typing import TypeVar, Type, final


__all__ = 'Singleton',


_T = TypeVar('_T', bound='Singleton')


class Singleton(ABC):
    __instance = None

    @final
    def __new__(cls, *args, **kwargs):  # type: ignore
        raise SyntaxWarning(f'{cls.__name__} is a singleton!')

    @final
    @classmethod
    def get_instance(cls: Type[_T]) -> _T:
        assert cls is not Singleton and issubclass(cls, Singleton)
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
            cls.__instance.__init__()
        return cls.__instance



import os
from typing import Iterable


from .io import DanmakuReader, DanmakuWriter
from .combination import DanmakuCombinator, TimebaseType
from .concatenation import DanmakuConcatenator


async def has_danmu(path: str) -> bool:
    async with DanmakuReader(path) as reader:
        async for _ in reader.read_danmus():
            return True
        return False


async def clear_danmu(path: str) -> None:
    async with DanmakuReader(path) as reader:
        async with DanmakuWriter(path) as writer:
            metadata = await reader.read_metadata()
            await writer.write_metadata(metadata)


async def concat_danmaku(
    in_paths: Iterable[str], deltas: Iterable[int], out_path: str
) -> None:
    concatenator = DanmakuConcatenator(in_paths, deltas, out_path)
    await concatenator.concat()


async def combine_danmaku(
    in_paths: Iterable[str],
    out_path: str,
    timebase_type: TimebaseType = TimebaseType.RECORD,
) -> None:
    combiator = DanmakuCombinator(in_paths, out_path, timebase_type)
    await combiator.combine()


async def merge_danmaku(src: str, dst: str, insert: bool = False) -> None:
    """Merge all danmus in the src into the dst.

    insert: if true, src insert before dst otherwise src append after dst
    """
    out = dst + '.tmp'
    if insert:
        await combine_danmaku([src, dst], out)
    else:
        await combine_danmaku([dst, src], out)
    await clear_danmu(src)
    os.replace(out, dst)



import asyncio
import os
import threading
from abc import ABC, abstractmethod
from typing import Awaitable, TypeVar, final

from blrec.logging.room_id import aio_task_with_room_id


class SwitchableMixin(ABC):
    def __init__(self) -> None:
        super().__init__()
        self._enabled = False
        self._enabled_lock = threading.Lock()

    @property
    def enabled(self) -> bool:
        with self._enabled_lock:
            return self._enabled

    @final
    def enable(self) -> None:
        with self._enabled_lock:
            if self._enabled:
                return
            self._enabled = True
            self._do_enable()

    @final
    def disable(self) -> None:
        with self._enabled_lock:
            if not self._enabled:
                return
            self._enabled = False
            self._do_disable()

    @abstractmethod
    def _do_enable(self) -> None:
        ...

    @abstractmethod
    def _do_disable(self) -> None:
        ...


class StoppableMixin(ABC):
    def __init__(self) -> None:
        super().__init__()
        self._stopped = True
        self._stopped_lock = threading.Lock()

    @property
    def stopped(self) -> bool:
        with self._stopped_lock:
            return self._stopped

    @final
    def start(self) -> None:
        with self._stopped_lock:
            if not self._stopped:
                return
            self._stopped = False
            self._do_start()

    @final
    def stop(self) -> None:
        with self._stopped_lock:
            if self._stopped:
                return
            self._stopped = True
            self._do_stop()

    @abstractmethod
    def _do_start(self) -> None:
        ...

    @abstractmethod
    def _do_stop(self) -> None:
        ...


class AsyncStoppableMixin(ABC):
    def __init__(self) -> None:
        super().__init__()
        self._stopped = True
        self._stopped_lock = asyncio.Lock()

    @property
    def stopped(self) -> bool:
        return self._stopped

    @final
    async def start(self) -> None:
        async with self._stopped_lock:
            if not self._stopped:
                return
            self._stopped = False
            await self._do_start()

    @final
    async def stop(self) -> None:
        async with self._stopped_lock:
            if self._stopped:
                return
            self._stopped = True
            await self._do_stop()

    @abstractmethod
    async def _do_start(self) -> None:
        ...

    @abstractmethod
    async def _do_stop(self) -> None:
        ...


_T = TypeVar('_T')


class AsyncCooperationMixin(ABC):
    def __init__(self) -> None:
        super().__init__()
        self._loop = asyncio.get_running_loop()

    def _submit_exception(self, exc: BaseException) -> None:
        from ..exception import submit_exception

        async def wrapper() -> None:
            # call submit_exception in a coroutine
            # workaround for `RuntimeError: no running event loop`
            submit_exception(exc)

        self._run_coroutine(wrapper())

    def _run_coroutine(self, coro: Awaitable[_T]) -> _T:
        future = asyncio.run_coroutine_threadsafe(self._with_room_id(coro), self._loop)
        return future.result()

    @aio_task_with_room_id
    async def _with_room_id(self, coro: Awaitable[_T]) -> _T:
        return await coro


class SupportDebugMixin(ABC):
    def __init__(self) -> None:
        super().__init__()

    def _init_for_debug(self, room_id: int) -> None:
        if (value := os.environ.get('BLREC_DEBUG')) and (
            value == '*' or room_id in value.split(',')
        ):
            self._debug = True
            self._debug_dir = os.path.expanduser(f'~/.blrec/debug/{room_id}')
            self._debug_dir = os.path.normpath(self._debug_dir)
            os.makedirs(self._debug_dir, exist_ok=True)
        else:
            self._debug = False
            self._debug_dir = ''



from reactivex import Observable, Subject

from ..utils.patterns import Singleton

__all__ = ('ExceptionCenter',)


class ExceptionCenter(Singleton):
    def __init__(self) -> None:
        super().__init__()
        self._source: Subject[BaseException] = Subject()

    @property
    def exceptions(self) -> Observable[BaseException]:
        return self._source

    def submit(self, exc: BaseException) -> None:
        self._source.on_next(exc)



import json
import logging
import os

import attr
from reactivex import Observable
from reactivex import operators as ops

from ..path import extra_metadata_path
from . import operators as flv_ops
from .operators.helpers import from_file

__all__ = 'AnalysingProgress', 'analyse_metadata'


logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class AnalysingProgress:
    count: int
    total: int


def analyse_metadata(
    path: str, *, display_progress: bool = False
) -> Observable[AnalysingProgress]:
    filesize = os.path.getsize(path)
    filename = os.path.basename(path)

    def dump_metadata() -> None:
        try:
            metadata = analyser.make_metadata()
            data = attr.asdict(metadata, filter=lambda a, v: v is not None)
            file_path = extra_metadata_path(path)
            with open(file_path, 'wt', encoding='utf8') as file:
                json.dump(data, file)
        except Exception as e:
            logger.error(f'Failed to dump metadata: {e}')
        else:
            logger.debug(f"Successfully dumped metadata to file: '{path}'")

    analyser = flv_ops.Analyser()
    return from_file(path).pipe(
        analyser,
        flv_ops.ProgressBar(
            desc='Analysing',
            postfix=filename,
            total=filesize,
            disable=not display_progress,
        ),
        ops.map(lambda i: len(i)),  # type: ignore
        ops.scan(lambda acc, x: acc + x, 0),  # type: ignore
        ops.map(lambda s: AnalysingProgress(s, filesize)),  # type: ignore
        ops.do_action(on_completed=dump_metadata),
    )



from __future__ import annotations

import asyncio
import logging
import time
from typing import Optional, TypeVar

import aiohttp
import requests
import urllib3
from reactivex import Observable, abc

from blrec.utils import operators as utils_ops

__all__ = ('RequestExceptionHandler',)


logger = logging.getLogger(__name__)

_T = TypeVar('_T')


class RequestExceptionHandler:
    def __init__(self) -> None:
        self._last_retry_time = time.monotonic()

    def __call__(self, source: Observable[_T]) -> Observable[_T]:
        return self._handle(source).pipe(
            utils_ops.retry(should_retry=self._should_retry)
        )

    def _handle(self, source: Observable[_T]) -> Observable[_T]:
        def subscribe(
            observer: abc.ObserverBase[_T],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            def on_error(exc: Exception) -> None:
                try:
                    raise exc
                except requests.exceptions.RequestException:  # XXX: ConnectionError
                    logger.warning(repr(exc))
                except urllib3.exceptions.HTTPError:
                    logger.warning(repr(exc))
                except asyncio.exceptions.TimeoutError:
                    logger.warning(repr(exc))
                except aiohttp.ClientError:
                    logger.warning(repr(exc))
                except Exception:
                    pass

                if self._should_retry(exc):
                    if time.monotonic() - self._last_retry_time < 1:
                        time.sleep(1)
                    self._last_retry_time = time.monotonic()

                observer.on_error(exc)

            return source.subscribe(
                observer.on_next, on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe)

    def _should_retry(self, exc: Exception) -> bool:
        if isinstance(
            exc,
            (
                requests.exceptions.RequestException,  # XXX: ConnectionError
                urllib3.exceptions.HTTPError,
                asyncio.exceptions.TimeoutError,
                aiohttp.ClientError,
            ),
        ):
            return True
        else:
            return False



from __future__ import annotations

import io
import logging
import os
from typing import Optional, Union

import av
from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

from blrec.bili.live import Live

from .segment_fetcher import InitSectionData, SegmentData

__all__ = ('SegmentRemuxer',)


logger = logging.getLogger(__name__)

TRACE_REMUX_SEGMENT = bool(os.environ.get('BLREC_TRACE_REMUX_SEGMENT'))
TRACE_LIBAV = bool(os.environ.get('BLREC_TRACE_LIBAV'))
if TRACE_LIBAV:
    logging.getLogger('libav').setLevel(5)
else:
    av.logging.set_level(av.logging.FATAL)


class SegmentRemuxer:
    def __init__(self, live: Live) -> None:
        self._live = live

    def __call__(
        self, source: Observable[Union[InitSectionData, SegmentData]]
    ) -> Observable[bytes]:
        return self._remux(source)

    def _remux(
        self, source: Observable[Union[InitSectionData, SegmentData]]
    ) -> Observable[bytes]:
        def subscribe(
            observer: abc.ObserverBase[bytes],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()

            init_section_data: Optional[bytes] = None

            def reset() -> None:
                nonlocal init_section_data
                init_section_data = None

            def on_next(data: Union[InitSectionData, SegmentData]) -> None:
                nonlocal init_section_data

                if isinstance(data, InitSectionData):
                    init_section_data = data.payload
                    observer.on_next(b'')
                    return

                if init_section_data is None:
                    return

                try:
                    remuxed_data = self._remux_segemnt(init_section_data + data.payload)
                except av.FFmpegError as e:
                    logger.warning(f'Failed to remux segment: {repr(e)}', exc_info=e)
                else:
                    observer.on_next(remuxed_data)

            def dispose() -> None:
                nonlocal disposed
                disposed = True
                reset()

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    def _remux_segemnt(self, data: bytes, format: str = 'flv') -> bytes:
        in_file = io.BytesIO(data)
        out_file = io.BytesIO()

        with av.open(in_file) as in_container:
            with av.open(out_file, mode='w', format=format) as out_container:
                in_video_stream = in_container.streams.video[0]
                in_audio_stream = in_container.streams.audio[0]
                out_video_stream = out_container.add_stream(template=in_video_stream)
                out_audio_stream = out_container.add_stream(template=in_audio_stream)

                for packet in in_container.demux():
                    if TRACE_REMUX_SEGMENT:
                        logger.debug(repr(packet))
                    # We need to skip the "flushing" packets that `demux` generates.
                    if packet.dts is None:
                        continue
                    # We need to assign the packet to the new stream.
                    if packet.stream.type == 'video':
                        packet.stream = out_video_stream
                    elif packet.stream.type == 'audio':
                        packet.stream = out_audio_stream
                    else:
                        raise NotImplementedError(packet.stream.type)
                    out_container.mux(packet)

        return out_file.getvalue()
